/*****************************************************************************
 * This software module was originally developed by Monotype Imaging Inc. 
 * in the course of development of the ISO/IEC 14496-22 "Open Font Format" 
 * for reference purposes and its performance may not have been optimized. 
 * This software module is an implementation of one or more tools as 
 * specified by the ISO/IEC 14496-22 "Open Font Format".
 * ISO/IEC gives You a royalty-free, worldwide, non-exclusive, copyright 
 * license to copy, distribute, and make derivative works of this software 
 * module or modifications thereof for use in implementations of the 
 * ISO/IEC 14496-22 in products that satisfy conformance criteria (if any). 
 * Those intending to use this software module in products are advised that 
 * its use may infringe existing patents. ISO/IEC have no liability for use 
 * of this software module or modifications thereof.
 * Copyright is not released for products that do not conform to audiovisual 
 * and image-coding related ITU Recommendations and/or ISO/IEC International 
 * Standards. 
 *
 * Monotype Imaging Inc. grants ISO/IEC all rights necessary to include the 
 * originally developed software module or modifications thereof in the 
 * ISO/IEC 14496-22 and to permit ISO/IEC to offer You a royalty-free, 
 * worldwide, non-exclusive, copyright license to copy, distribute, and make 
 * derivative works for use in implementations of the ISO/IEC 14496-22 in 
 * products that satisfy conformance criteria (if any), and to the extent 
 * that such originally developed software module or portions of it are 
 * included in the ISO/IEC 14496-22. To the extent that Monotype Imaging Inc. 
 * owns patent rights that would be required to make, use, or sell the 
 * originally developed software module or portions thereof included in the 
 * ISO/IEC 14496-22 in a conforming product, Monotype Imaging Inc. will 
 * assure the ISO/IEC that it is willing to negotiate licenses under 
 * reasonable and non-discriminatory terms and conditions with applicants 
 * throughout the world.
 *
 * ISO/IEC gives You a free license to this software module or modifications 
 * thereof for the sole purpose of developing the ISO/IEC 14496-22 "Open Font 
 * Format".
 *
 * Monotype Imaging Inc. retains full right to modify and use the code for 
 * its own purpose, assign or donate the code to a third party and to inhibit 
 * third parties from using the code for products that do not conform to 
 * MPEG-related ITU Recommendations and/or ISO/IEC International Standards. 
 *
 * This copyright notice must be included in all copies or derivative works. 
 * Copyright (c) ISO/IEC 2007. 
 ******************************************************************************/

#include <stdio.h>
#include <time.h>
#include <malloc.h>

#define BIT0 0x01
#define BIT1 0x02
#define BIT2 0x04
#define BIT3 0x08
#define BIT4 0x10
#define BIT5 0x20
#define BIT6 0x40
#define BIT7 0x80
#define BIT8 0x100

/* non  errors */
#define FILECHECKSUMERROR	1

#define CHECKSUMERRORHEAD	2
#define CHECKSUMERRORHHEA	3
#define CHECKSUMERRORMAXP	4
#define CHECKSUMERROROS2	5
#define CHECKSUMERRORNAME	6
#define CHECKSUMERRORPOST	7
#define CHECKSUMERRORHMTX	8
#define CHECKSUMERRORCMAP	9
#define CHECKSUMERRORPREP	10
#define CHECKSUMERRORFPGM	11
#define CHECKSUMERRORCVT	12
#define CHECKSUMERRORVORG	13
#define CHECKSUMERRORCFF	14
#define CHECKSUMERROREBLC	15
#define CHECKSUMERROREBDT	16
#define CHECKSUMERROREBSC	17
#define CHECKSUMERRORBASE	18
#define CHECKSUMERRORGDEF	19
#define CHECKSUMERRORGSUB	20
#define CHECKSUMERRORGPOS	21
#define CHECKSUMERRORJSTF	22	
#define CHECKSUMERRORLOCA	23
#define CHECKSUMERRORGLYF	24	

/*  errors */
#define ERRORINFILEHEADER	25

#define ERRORINTABLEHEAD	26
#define ERRORINTABLEHHEA	27
#define ERRORINTABLEMAXP	28
#define ERRORINTABLEOS2		29
#define ERRORINTABLENAME	30
#define ERRORINTABLEPOST	31
#define ERRORINTABLEHMTX	32
#define ERRORINTABLECMAP	33
#define ERRORINTABLEPREP	34
#define ERRORINTABLEFPGM	35
#define ERRORINTABLECVT		36
#define ERRORINTABLEVORG	37
#define ERRORINTABLECFF		38
#define ERRORINTABLEEBLC	39
#define ERRORINTABLEEBDT	40
#define ERRORINTABLEEBSC	41
#define ERRORINTABLEBASE	42
#define ERRORINTABLEGDEF	43
#define ERRORINTABLEGSUB	44
#define ERRORINTABLEGPOS	45
#define ERRORINTABLEJSTF	46
#define ERRORINTABLELOCA	47
#define ERRORINTABLEGLYF	48

#define TABLENOTFOUNDHEAD	49
#define TABLENOTFOUNDHHEA	50
#define TABLENOTFOUNDMAXP	51
#define TABLENOTFOUNDOS2	52
#define TABLENOTFOUNDNAME	53
#define TABLENOTFOUNDPOST	54
#define TABLENOTFOUNDHMTX	55
#define TABLENOTFOUNDCMAP	56
#define TABLENOTFOUNDVORG	57
#define TABLENOTFOUNDCFF	58
#define TABLENOTFOUNDJSTF	59	
#define TABLENOTFOUNDLOCA	60
#define TABLENOTFOUNDGLYF	61	

char *errorstring[62] = {
	"ERROR",
	"File Checksum Error",
	"Checksum Error in head table",
	"Checksum Error in hhea table",
	"Checksum Error in maxp table",
	"Checksum Error in OS/2 table",
	"Checksum Error in name table",
	"Checksum Error in post table",
	"Checksum Error in hmtx table",
	"Checksum Error in cmap table",
	"Checksum Error in prep table",
	"Checksum Error in fpgm table",
	"Checksum Error in cvt table",
	"Checksum Error in VORG table",
	"Checksum Error in CFF table",
	"Checksum Error in EBLC table",
	"Checksum Error in EBDT table",
	"Checksum Error in EBSC table",
	"Checksum Error in BASE table",
	"Checksum Error in GDEF table",
	"Checksum Error in GSUB table",
	"Checksum Error in GPOS table",
	"Checksum Error in JSTF table",
	"Checksum Error in LOCA table",
	"Checksum Error in GLYF table",
	"Error in file header",
	"Error in head table",
	"Error in hhea table",
	"Error in maxp table",
	"Error in OS/2 table",
	"Error in name table",
	"Error in post table",
	"Error in hmtx table",
	"Error in cmap table",
	"Error in prep table",
	"Error in fpgm table",
	"Error in cvt table",
	"Error in VORG table",
	"Error in CFF table",
	"Error in EBLC table",
	"Error in EBDT table",
	"Error in EBSC table",
	"Error in BASE table",
	"Error in GDEF table",
	"Error in GSUB table",
	"Error in GPOS table",
	"Error in JSTF table",
	"Error in loca table",
	"Error in glyph table",
	"head table not found",
	"hhea table not found",
	"maxp table not found",
	"OS/2 table not found",
	"name table not found",
	"post table not found",
	"hmtx table not found",
	"cmap table not found",
	"VORG table not found",
	"CFF table not found",
	"JSTF table not found",
	"loca table not found",
	"glyf table not found"};
	


typedef struct {
  unsigned long tag;
  unsigned long checksum;
  unsigned long offset;
  unsigned long size;
} table_type;

typedef struct {
	long version;
	short numtables;
	short searchRange;
	short entrySelector;
	short rangeShift;
	table_type *tables;
	int numGlyphs;
	int indexToLocFormat;
	int numberOfHMetrics;
	int numberOfErrors;
	int errors[50];
	int level1;
	int level2;
	int level3;
	int level4;
	unsigned int filelen;
} truetype_type;

table_type *gettable(truetype_type *tt,char *tag);
int verifytruetype(unsigned char *buf, truetype_type *tt, int len);

int verifyttc(unsigned char *buf, unsigned char *ptr, truetype_type *tt, int len);
int verifyttfotf(unsigned char *buf, unsigned char *ptr, truetype_type *tt, int len);
int verifygpostable(unsigned char *buf, truetype_type *tt);
int verifygsubtable(unsigned char *buf, truetype_type *tt);
int verifyscriptlist(unsigned int tableoffset,unsigned char *buf, table_type *tptr);
int verifylangsystable(unsigned int tableoffset,unsigned char *buf, table_type *tptr);
int verifyfeaturelist(unsigned int tableoffset,unsigned char *buf, table_type *tptr);
int verifygsublookuplist(unsigned int tableoffset,unsigned char *buf, table_type *tptr);
int verifygsublookup8(unsigned int tableoffset,unsigned char *buf, table_type *tptr);
int verifygsublookup7(unsigned int tableoffset,unsigned char *buf, table_type *tptr);
int verifygsublookup6(unsigned int tableoffset,unsigned char *buf, table_type *tptr);
int verifygsublookup5(unsigned int tableoffset,unsigned char *buf, table_type *tptr);
int verifygsublookup4(unsigned int tableoffset,unsigned char *buf, table_type *tptr);
int verifygsublookup3(unsigned int tableoffset,unsigned char *buf, table_type *tptr);
int verifygsublookup2(unsigned int tableoffset,unsigned char *buf, table_type *tptr);
int verifygsublookup1(unsigned int tableoffset,unsigned char *buf, table_type *tptr);
int verifygposlookuplist(unsigned int tableoffset,unsigned char *buf, table_type *tptr);
int verifygposlookup9(unsigned int tableoffset,unsigned char *buf, table_type *tptr);
int verifygposlookup8(unsigned int tableoffset,unsigned char *buf, table_type *tptr);
int verifygposlookup7(unsigned int tableoffset,unsigned char *buf, table_type *tptr);
int verifygposlookup6(unsigned int tableoffset,unsigned char *buf, table_type *tptr);
int verifygposlookup5(unsigned int tableoffset,unsigned char *buf, table_type *tptr);
int verifygposlookup4(unsigned int tableoffset,unsigned char *buf, table_type *tptr);
int verifygposlookup3(unsigned int tableoffset,unsigned char *buf, table_type *tptr);
int verifygposlookup2(unsigned int tableoffset,unsigned char *buf, table_type *tptr);
int verifygposlookup1(unsigned int tableoffset,unsigned char *buf, table_type *tptr);
int verifymarkarray(unsigned int tableoffset,unsigned char *buf, table_type *tptr);
int verifyanchortable(unsigned int tableoffset,unsigned char *buf, table_type *tptr);
int getvaluerecordinfo(unsigned int format, unsigned char *ptr, unsigned int *len, 
					   unsigned int *offset1, unsigned int *offset2,
					   unsigned int *offset3, unsigned int *offset4,
					   unsigned char *buf, table_type *tptr);
int verifyjstftable(unsigned char *buf, truetype_type *tt);
int verifybasetable(unsigned char *buf, truetype_type *tt);
int verifyaxistable(unsigned int tableoffset,unsigned char *buf, table_type *tptr);
int verifyminmaxtable(unsigned int tableoffset,unsigned char *buf, table_type *tptr);
int verifybasecoordtable(unsigned int tableoffset,unsigned char *buf, table_type *tptr);
int verifyjstflangsystable(char *ptr,char *buf,table_type *tptr);
int verifygdeftable(unsigned char *buf, truetype_type *tt);
int verifyattachlisttable(unsigned int tableoffset,unsigned char *buf, table_type *tptr);
int verifyligcaretlisttable(unsigned int tableoffset,unsigned char *buf, table_type *tptr);
int verifycoveragetable(unsigned int tableoffset,unsigned char *buf, table_type *tptr);
int verifyclassdeftable(unsigned int tableoffset,unsigned char *buf, table_type *tptr);
int verifydevicetable(unsigned int tableoffset,unsigned char *buf, table_type *tptr);
int verifyeblcebdttable(unsigned char *buf, truetype_type *tt);
int verifyebsctable(unsigned char *buf, truetype_type *tt);
int verifyglyftable(unsigned char *buf, truetype_type *tt);
int verifysimpleglyph(char *buf, int size);
int verifycompositeglyph(char *buf, int size);
int verifyposttable(unsigned char *buf, truetype_type *tt);
int verifylocatable(unsigned char *buf, truetype_type *tt);
int verifyhmtxtable(unsigned char *buf, truetype_type *tt);
int verifycfftable(unsigned char *buf, truetype_type *tt);
int verifyvorgtable(unsigned char *buf, truetype_type *tt);
int verifyfpgmtable(unsigned char *buf, truetype_type *tt);
int verifypreptable(unsigned char *buf, truetype_type *tt);
int verifypreptable(unsigned char *buf, truetype_type *tt);
int verifycvttable(unsigned char *buf, truetype_type *tt);
int verifycmaptable(unsigned char *buf, truetype_type *tt);
int verifycmapformat12(unsigned char *ptr, unsigned char *buf, table_type *tptr);
int verifycmapformat10(unsigned char *ptr, unsigned char *buf, table_type *tptr);
int verifycmapformat8(unsigned char *ptr, unsigned char *buf, table_type *tptr);
int verifycmapformat6(unsigned char *ptr, unsigned char *buf, table_type *tptr);
int verifycmapformat4(unsigned char *ptr, unsigned char *buf, table_type *tptr);
int verifycmapformat2(unsigned char *ptr, unsigned char *buf, table_type *tptr);
int verifycmapformat0(unsigned char *ptr, unsigned char *buf, table_type *tptr);
int verifynametable(unsigned char *buf, truetype_type *tt);
int verifyos2table(unsigned char *buf, truetype_type *tt);
int verifymaxptable(unsigned char *buf, truetype_type *tt);
int verifyhheatable(unsigned char *buf, truetype_type *tt);
int verifyheadtable(unsigned char *buf, truetype_type *tt);
int istable(table_type *t,char *tag);



table_type *gettable(truetype_type *tt,char *tag)
{
  table_type *tptr = tt->tables;
  
  int i;
  
  for(i=0; i<tt->numtables; i++) {
    if(tptr->tag == *(unsigned long *)tag) {
      return tptr;
      
    }
    tptr++;
  }

  return NULL;
}

int istable(table_type *t,char *tag)
{
  
  
  if(t->tag == *(unsigned long *)tag) 
      return 1;
  else
     return 0;
}


int verifyheadtable(unsigned char *buf, truetype_type *tt)
{
	table_type *tptr = gettable(tt,"head");
	int i;
	unsigned char *ptr;
	unsigned long checksum = 0;
	long magicNumber;
	

	if(!tptr) {
		tt->errors[tt->numberOfErrors] = TABLENOTFOUNDHEAD;
		tt->numberOfErrors++;
		tt->level1 = 0;
		return -1;
	}

	ptr=buf+tptr->offset;

	if(tptr->offset + tptr->size > tt->filelen)	{
		tt->errors[tt->numberOfErrors] = ERRORINTABLEHEAD;
		tt->numberOfErrors++;
		tt->level1 = 0;
		return -1;
	}  

	for(i=0; (unsigned long)i<(tptr->size+3)>>2; i++) {

		/* skip checksumadjustment */
		if(i == 2) {
			ptr += 4;
			continue; 
		}
		
		checksum += (*ptr++)<<24;
		checksum += (*ptr++)<<16;
		checksum += (*ptr++)<<8;
		checksum += (*ptr++);
		
	}

	

	if(checksum != tptr->checksum) {
		tt->errors[tt->numberOfErrors] = CHECKSUMERRORHEAD;
		tt->numberOfErrors++;
	}

	if(tptr->size != 54) {
		tt->errors[tt->numberOfErrors] = ERRORINTABLEHEAD;
		tt->numberOfErrors++;
		tt->level1 = 0;
	}

	ptr = buf + tptr->offset + 12;
	magicNumber = (*ptr++)<<24;
	magicNumber += (*ptr++)<<16;
	magicNumber += (*ptr++)<<8;
	magicNumber += (*ptr++);

	if(magicNumber != 0x5f0f3cf5) {
		tt->errors[tt->numberOfErrors] = ERRORINTABLEHEAD;
		tt->numberOfErrors++;

	}

	ptr = buf + tptr->offset + 50;
	tt->indexToLocFormat = (*ptr++)<<8;
	tt->indexToLocFormat += (*ptr++);

	return 0;
}

int verifyhheatable(unsigned char *buf, truetype_type *tt)
{
	table_type *tptr = gettable(tt,"hhea");
	int i;
	unsigned char *ptr;
	unsigned long checksum = 0;

	if(!tptr) {
		tt->errors[tt->numberOfErrors] = TABLENOTFOUNDHHEA;
		tt->numberOfErrors++;
		tt->level1 = 0;
		return -1;
	}

	ptr=buf+tptr->offset;

	if(tptr->offset + tptr->size > tt->filelen)	{
		tt->errors[tt->numberOfErrors] = ERRORINTABLEHHEA;
		tt->numberOfErrors++;
		tt->level1 = 0;
		return -1;
	}  

	for(i=0; (unsigned long)i<(tptr->size+3)>>2; i++) {
		
		checksum += (*ptr++)<<24;
		checksum += (*ptr++)<<16;
		checksum += (*ptr++)<<8;
		checksum += (*ptr++);
		
	}

	if(checksum != tptr->checksum) {
		tt->errors[tt->numberOfErrors] = CHECKSUMERRORHHEA;
		tt->numberOfErrors++;
	}

	if(tptr->size != 36) {
		tt->errors[tt->numberOfErrors] = ERRORINTABLEHHEA;
		tt->numberOfErrors++;
		tt->level1 = 0;
	}

	
	
	ptr = buf + tptr->offset + 34;
	tt->numberOfHMetrics = (*ptr++)<<8;
	tt->numberOfHMetrics += (*ptr++);

	return 0;
}

int verifymaxptable(unsigned char *buf, truetype_type *tt)
{
	table_type *tptr = gettable(tt,"maxp");
	int i;
	unsigned char *ptr;
	unsigned long checksum = 0;
	unsigned long version;
	unsigned long otto = 0x4f54544f;

	if(!tptr) {
		tt->errors[tt->numberOfErrors] = TABLENOTFOUNDMAXP;
		tt->numberOfErrors++;
		tt->level1 = 0;
		return -1;
	}

	ptr=buf+tptr->offset;

	if(tptr->offset + tptr->size > tt->filelen)	{
		tt->errors[tt->numberOfErrors] = ERRORINTABLEMAXP;
		tt->numberOfErrors++;
		tt->level1 = 0;
		return -1;
	}  

	for(i=0; (unsigned long)i<(tptr->size+3)>>2; i++) {
		
		checksum += (*ptr++)<<24;
		checksum += (*ptr++)<<16;
		checksum += (*ptr++)<<8;
		checksum += (*ptr++);
		
	}

	if(checksum != tptr->checksum) {
		tt->errors[tt->numberOfErrors] = CHECKSUMERRORMAXP;
		tt->numberOfErrors++;
	}

	version = (*ptr++)<<24;
	version += (*ptr++)<<16;
	version += (*ptr++)<<8;
	version += (*ptr++);


	if(version == 0x10000 && tptr->size != 32) {
		tt->errors[tt->numberOfErrors] = ERRORINTABLEMAXP;
		tt->numberOfErrors++;
		tt->level1 = 0;
		return -1;
	}

	if((version == otto || version == 0x5000) && tptr->size != 6) {
		tt->errors[tt->numberOfErrors] = ERRORINTABLEMAXP;
		tt->numberOfErrors++;
		tt->level1 = 0;
		return -1;
	}

	
	
	ptr = buf + tptr->offset + 4;
	tt->numGlyphs = (*ptr++)<<8;
	tt->numGlyphs += (*ptr++);

	return 0;
}

int verifyos2table(unsigned char *buf, truetype_type *tt)
{
	table_type *tptr = gettable(tt,"OS/2");
	int i;
	unsigned char *ptr;
	unsigned long checksum = 0;

	if(!tptr) {
		tt->errors[tt->numberOfErrors] = TABLENOTFOUNDOS2;
		tt->numberOfErrors++;
		tt->level1 = 0;
		return 0;
	}

	ptr=buf+tptr->offset;

	if(tptr->offset + tptr->size > tt->filelen)	{
		tt->errors[tt->numberOfErrors] = ERRORINTABLEOS2;
		tt->numberOfErrors++;
		tt->level1 = 0;
		return 0;
	}  

	for(i=0; (unsigned long)i<(tptr->size+3)>>2; i++) {
		
		checksum += (*ptr++)<<24;
		checksum += (*ptr++)<<16;
		checksum += (*ptr++)<<8;
		checksum += (*ptr++);
		
	}

	if(checksum != tptr->checksum) {
		tt->errors[tt->numberOfErrors] = CHECKSUMERROROS2;
		tt->numberOfErrors++;
	}

	if(tptr->size < 78) {
		tt->errors[tt->numberOfErrors] = ERRORINTABLEOS2;
		tt->numberOfErrors++;
		tt->level1 = 0;
	}

	


	return 0;
}
int verifynametable(unsigned char *buf, truetype_type *tt)
{
	table_type *tptr = gettable(tt,"name");
	unsigned int i;
	unsigned char *ptr;
	unsigned long checksum = 0;
	unsigned int count;
	unsigned int stringoffset;
	unsigned int stringlen;
	unsigned int offset;

	if(!tptr) {
		tt->errors[tt->numberOfErrors] = TABLENOTFOUNDNAME;
		tt->numberOfErrors++;
		tt->level1 = 0;
		return 0;
	}

	ptr=buf+tptr->offset;

	if(tptr->offset + tptr->size > tt->filelen)	{
		tt->errors[tt->numberOfErrors] = ERRORINTABLENAME;
		tt->numberOfErrors++;
		tt->level1 = 0;
		return 0;
	}  
	

	for(i=0; i<(tptr->size+3)>>2; i++) {
		
		checksum += (*ptr++)<<24;
		checksum += (*ptr++)<<16;
		checksum += (*ptr++)<<8;
		checksum += (*ptr++);
		
	}

	if(checksum != tptr->checksum) {
		tt->errors[tt->numberOfErrors] = CHECKSUMERRORNAME;
		tt->numberOfErrors++;
	}

	
	ptr = buf + tptr->offset + 2;

	if(ptr+4 > buf+tptr->offset+tptr->size){
		tt->errors[tt->numberOfErrors] = ERRORINTABLENAME;
		tt->numberOfErrors++;
		tt->level1 = 0;
		return 0;
	}
		
	count = (*ptr++)<<8;
	count += (*ptr++);

	stringoffset = (*ptr++)<<8;
	stringoffset += (*ptr++);

	if(count * 12 + 6 > tptr->size) {
		tt->errors[tt->numberOfErrors] = ERRORINTABLENAME;
		tt->numberOfErrors++;
		tt->level1 = 0;
		return 0;
	}

	
	for(i=0; i<count; i++) {
		ptr += 8;
		if(ptr+4 > buf+tptr->offset+tptr->size){
			tt->errors[tt->numberOfErrors] = ERRORINTABLENAME;
			tt->numberOfErrors++;
			return 0;
		} 

		stringlen = (*ptr++)<<8;
		stringlen += (*ptr++);

		offset = (*ptr++)<<8;
		offset += (*ptr++);

		if(stringoffset + offset + stringlen > tptr->size) {
			tt->errors[tt->numberOfErrors] = ERRORINTABLENAME;
			tt->numberOfErrors++;
			tt->level1 = 0;
			return 0;
		} 
	  
	} 


 
	return 0;
}
int verifycmapformat0(unsigned char *ptr, unsigned char *buf, table_type *tptr)
{
	unsigned int length;

	if(ptr+2 > buf+tptr->offset+tptr->size)
		return -1;
	
	length = (*ptr++)<<8;
	length += (*ptr++);

	if(length > 256+6)
		return -1;

	if(ptr+length-4 > buf+tptr->offset+tptr->size)
		return -1;


	return 0;
}
int verifycmapformat2(unsigned char *ptr, unsigned char *buf, table_type *tptr)
{
	unsigned int length;
	
	if(ptr+2 > buf+tptr->offset+tptr->size)
		return -1;

	length = (*ptr++)<<8;
	length += (*ptr++);

	if(ptr+length-4 > buf+tptr->offset+tptr->size)
		return -1;

	return 0;
}
int verifycmapformat4(unsigned char *ptr, unsigned char *buf, table_type *tptr)
{
	unsigned int length;
	unsigned int segcountx2;
	
	if(ptr+6 > buf+tptr->offset+tptr->size)
		return -1;
	
	length = (*ptr++)<<8;
	length += (*ptr++);

	ptr += 2;

	segcountx2 = (*ptr++)<<8;
	segcountx2 += (*ptr++);

	if(4*segcountx2+16 > length)
		return -1;

	if(ptr +length - 8 > buf+tptr->offset+tptr->size)
		return -1;

	return 0;
}
int verifycmapformat6(unsigned char *ptr, unsigned char *buf, table_type *tptr)
{
	unsigned int length;
	unsigned int count;

	if(ptr+6 > buf+tptr->offset+tptr->size)
		return -1;
	
	length = (*ptr++)<<8;
	length += (*ptr++);

	ptr += 4;

	count = (*ptr++)<<8;
	count += (*ptr++);

	if(ptr+count*2 > buf+tptr->offset+tptr->size)
		return -1;

	if(length < count*2+10)
		return -1;


	return 0;
}
int verifycmapformat8(unsigned char *ptr, unsigned char *buf, table_type *tptr)
{
	unsigned int length;
	unsigned int count;
	ptr += 2; /* skip reserved word */

	if(ptr+4 > buf+tptr->offset+tptr->size)
		return -1;

	length = (*ptr++)<<24;
	length += (*ptr++)<<16;
	length += (*ptr++)<<8;
	length += (*ptr++);

	ptr += 4 + 8192;

	if(ptr+4 > buf+tptr->offset+tptr->size)
		return -1;

	count = (*ptr++)<<24;
	count += (*ptr++)<<16;
	count += (*ptr++)<<8;
	count += (*ptr++);

	if(ptr+count*12 > buf+tptr->offset+tptr->size)
		return -1;

	if(length < count*12+8192+16)
		return -1;
	return 0;
}
int verifycmapformat10(unsigned char *ptr, unsigned char *buf, table_type *tptr)
{
	
	unsigned int length;
	unsigned int count;
	ptr += 2; /* skip reserved word */

	if(ptr+16 > buf+tptr->offset+tptr->size)
		return -1;

	length = (*ptr++)<<24;
	length += (*ptr++)<<16;
	length += (*ptr++)<<8;
	length += (*ptr++);

	ptr += 8;

	count = (*ptr++)<<24;
	count += (*ptr++)<<16;
	count += (*ptr++)<<8;
	count += (*ptr++);

	if(ptr+count*2 > buf+tptr->offset+tptr->size)
		return -1;

	if(length < count*2+20)
		return -1;


	return 0;
}
int verifycmapformat12(unsigned char *ptr, unsigned char *buf, table_type *tptr)
{
	unsigned int length;
	unsigned int count;
	ptr += 2; /* skip reserved word */
	
	if(ptr+12 > buf+tptr->offset+tptr->size)
		return -1;

	length = (*ptr++)<<24;
	length += (*ptr++)<<16;
	length += (*ptr++)<<8;
	length += (*ptr++);

	ptr += 4;

	count = (*ptr++)<<24;
	count += (*ptr++)<<16;
	count += (*ptr++)<<8;
	count += (*ptr++);

	if(ptr+count*12 > buf+tptr->offset+tptr->size)
		return -1;

	if(length < count*12+16)
		return -1;


	return 0;
}
int verifycmaptable(unsigned char *buf, truetype_type *tt)
{
	
	table_type *tptr = gettable(tt,"cmap");
	unsigned int i;
	unsigned char *ptr;
	unsigned long checksum = 0;
	unsigned int count;
	unsigned int offset;
	unsigned int format;
	unsigned char *ptr2;

	if(!tptr) {
		tt->errors[tt->numberOfErrors] = TABLENOTFOUNDCMAP;
		tt->numberOfErrors++;
		tt->level1 = 0;
		return 0;
	}

	ptr=buf+tptr->offset;

	if(tptr->offset + tptr->size > tt->filelen)	{
		tt->errors[tt->numberOfErrors] = ERRORINTABLECMAP;
		tt->numberOfErrors++;
		tt->level1 = 0;
		return 0;
	}  

	for(i=0; i<(tptr->size+3)>>2; i++) {
		
		checksum += (*ptr++)<<24;
		checksum += (*ptr++)<<16;
		checksum += (*ptr++)<<8;
		checksum += (*ptr++);
		
	}

	if(checksum != tptr->checksum) {
		tt->errors[tt->numberOfErrors] = CHECKSUMERRORCMAP;
		tt->numberOfErrors++;
	}


	ptr = buf + tptr->offset + 2;
	if(ptr+2 > buf+tptr->offset+tptr->size){
		tt->errors[tt->numberOfErrors] = ERRORINTABLECMAP;
		tt->numberOfErrors++;
		tt->level1 = 0;
		return 0;
	}
	count = (*ptr++)<<8;
	count += (*ptr++);

	for(i=0; i<count; i++) {
    
    
		ptr += 4;

		if(ptr+4 > buf+tptr->offset+tptr->size){
			tt->errors[tt->numberOfErrors] = ERRORINTABLECMAP;
			tt->numberOfErrors++;
			tt->level1 = 0;
			return 0;
		} 
    
		offset = (*ptr++)<<24;
		offset += (*ptr++)<<16;
		offset += (*ptr++)<<8;
		offset += (*ptr++);

		ptr2 = buf + tptr->offset + offset;

		if(ptr2+6 > buf+tptr->offset+tptr->size){
			tt->errors[tt->numberOfErrors] = ERRORINTABLECMAP;
			tt->numberOfErrors++;
			tt->level1 = 0;
			return 0;
		} 
  
	
		format = (*ptr2++)<<8;
		format += (*ptr2++);
	

		

		if(format == 0) {
			if(verifycmapformat0(ptr2,buf,tptr)) {
				tt->errors[tt->numberOfErrors] = ERRORINTABLECMAP;
				tt->numberOfErrors++;
				tt->level1 = 0;
				return 0;
			} 
		}

		else if(format == 2) {
			if(verifycmapformat2(ptr2,buf,tptr)) {
				tt->errors[tt->numberOfErrors] = ERRORINTABLECMAP;
				tt->numberOfErrors++;
				tt->level1 = 0;
				return 0;
			} 
		}
		else if(format == 4) {
			if(verifycmapformat4(ptr2,buf,tptr)) {
				tt->errors[tt->numberOfErrors] = ERRORINTABLECMAP;
				tt->numberOfErrors++;
				tt->level1 = 0;
				return 0;
			} 
		}
		else if(format == 6) {
			if(verifycmapformat6(ptr2,buf,tptr)) {
				tt->errors[tt->numberOfErrors] = ERRORINTABLECMAP;
				tt->numberOfErrors++;
				tt->level1 = 0;
				return 0;
			} 
		}
		else if(format == 8) {
			if(verifycmapformat8(ptr2,buf,tptr)) {
				tt->errors[tt->numberOfErrors] = ERRORINTABLECMAP;
				tt->numberOfErrors++;
				tt->level1 = 0;
				return 0;
			} 
		}
		else if(format == 10) {
			if(verifycmapformat10(ptr2,buf,tptr)) {
				tt->errors[tt->numberOfErrors] = ERRORINTABLECMAP;
				tt->numberOfErrors++;
				tt->level1 = 0;
				return 0;
			} 
		}
		else if(format == 12) {
			if(verifycmapformat12(ptr2,buf,tptr)) {
				tt->errors[tt->numberOfErrors] = ERRORINTABLECMAP;
				tt->numberOfErrors++;
				tt->level1 = 0;
				return 0;
			} 
		}
				

		
		else {
			tt->errors[tt->numberOfErrors] = ERRORINTABLECMAP;
			tt->numberOfErrors++;
			return 0;
		} 
		
	}

	return 0;
}

int verifycvttable(unsigned char *buf, truetype_type *tt)
{
	table_type *tptr = gettable(tt,"cvt ");
	unsigned int i;
	unsigned char *ptr;
	unsigned long checksum = 0;

	if(!tptr) {
		return 0;
	}
	
	ptr=buf+tptr->offset;

	if(tptr->offset + tptr->size > tt->filelen)	{
		tt->errors[tt->numberOfErrors] = ERRORINTABLECVT;
		tt->numberOfErrors++;
		return 0;
	}  

	for(i=0; i<(tptr->size+3)>>2; i++) {
		
		checksum += (*ptr++)<<24;
		checksum += (*ptr++)<<16;
		checksum += (*ptr++)<<8;
		checksum += (*ptr++);
		
	}

	if(checksum != tptr->checksum) {
		tt->errors[tt->numberOfErrors] = CHECKSUMERRORCVT;
		tt->numberOfErrors++;
	}

	return 0;


}

int verifypreptable(unsigned char *buf, truetype_type *tt)
{
	table_type *tptr = gettable(tt,"prep");
	unsigned int i;
	unsigned char *ptr;
	unsigned long checksum = 0;


	if(!tptr)
		return 0;

	if(tptr->offset + tptr->size > tt->filelen)	{
		tt->errors[tt->numberOfErrors] = ERRORINTABLEPREP;
		tt->numberOfErrors++;
		return 0;
	}  

	ptr=buf+tptr->offset;

	for(i=0; i<(tptr->size+3)>>2; i++) {
		
		checksum += (*ptr++)<<24;
		checksum += (*ptr++)<<16;
		checksum += (*ptr++)<<8;
		checksum += (*ptr++);
		
	}

	if(checksum != tptr->checksum) {
		tt->errors[tt->numberOfErrors] = CHECKSUMERRORPREP;
		tt->numberOfErrors++;
	}

	return 0;
}

int verifyfpgmtable(unsigned char *buf, truetype_type *tt)
{
	table_type *tptr = gettable(tt,"fpgm");
	unsigned int i;
	unsigned char *ptr;
	unsigned long checksum = 0;


	if(!tptr)
		return 0;

	if(tptr->offset + tptr->size > tt->filelen)	{
		tt->errors[tt->numberOfErrors] = ERRORINTABLEFPGM;
		tt->numberOfErrors++;
		return 0;
	}  

	ptr=buf+tptr->offset;

	for(i=0; i<(tptr->size+3)>>2; i++) {
		
		checksum += (*ptr++)<<24;
		checksum += (*ptr++)<<16;
		checksum += (*ptr++)<<8;
		checksum += (*ptr++);
		
	}

	if(checksum != tptr->checksum) {
		tt->errors[tt->numberOfErrors] = CHECKSUMERRORFPGM;
		tt->numberOfErrors++;
	}

	return 0;
}

int verifyvorgtable(unsigned char *buf, truetype_type *tt)
{
	table_type *tptr = gettable(tt,"VORG");
	unsigned int i;
	unsigned char *ptr;
	unsigned long checksum = 0;
	unsigned int count;


	if(!tptr)
		return 0;

	ptr=buf+tptr->offset;
	
	if(tptr->offset + tptr->size > tt->filelen)	{
		tt->errors[tt->numberOfErrors] = ERRORINTABLEVORG;
		tt->numberOfErrors++;
		return 0;
	}  

	for(i=0; i<(tptr->size+3)>>2; i++) {
		
		checksum += (*ptr++)<<24;
		checksum += (*ptr++)<<16;
		checksum += (*ptr++)<<8;
		checksum += (*ptr++);
		
	}

	if(checksum != tptr->checksum) {
		tt->errors[tt->numberOfErrors] = CHECKSUMERRORVORG;
		tt->numberOfErrors++;
	}

	ptr += 6;

	if(ptr+2 > buf+tptr->offset+tptr->size) {
		tt->errors[tt->numberOfErrors] = ERRORINTABLEVORG;
		tt->numberOfErrors++;
		return 0;
	}
	

	count = (*ptr++)<<8;
	count += (*ptr++);

	if(ptr + count * 4 > buf +tptr->offset+tptr->size) {
		tt->errors[tt->numberOfErrors] = ERRORINTABLEVORG;
		tt->numberOfErrors++;
		return 0;
	}
		

	return 0;
}


int verifycfftable(unsigned char *buf, truetype_type *tt)
{
	table_type *tptr = gettable(tt,"CFF ");
	unsigned int i;
	unsigned char *ptr;
	unsigned long checksum = 0;
	int otto = 0x4f54544f;

	if(!tptr && tt->version == otto) {
		tt->errors[tt->numberOfErrors] = TABLENOTFOUNDCFF;
		tt->numberOfErrors++;
		tt->level3 = 0;
		return 0;
	}
	if(!tptr ) {
		tt->level3 = 0;
		return 0;
	}
	ptr=buf+tptr->offset;

	if(tptr->offset + tptr->size > tt->filelen)	{
		tt->errors[tt->numberOfErrors] = ERRORINTABLECFF;
		tt->numberOfErrors++;
		tt->level3 = 0;
		return 0;
	}  

	for(i=0; i<(tptr->size+3)>>2; i++) {
		
		checksum += (*ptr++)<<24;
		checksum += (*ptr++)<<16;
		checksum += (*ptr++)<<8;
		checksum += (*ptr++);
		
	}

	if(checksum != tptr->checksum) {
		tt->errors[tt->numberOfErrors] = CHECKSUMERRORCFF;
		tt->numberOfErrors++;
	}

	return 0;
}


int verifyhmtxtable(unsigned char *buf, truetype_type *tt)
{
	table_type *tptr = gettable(tt,"hmtx");
	unsigned int i;
	unsigned char *ptr;
	unsigned long checksum = 0;

	if(!tptr) {
		tt->errors[tt->numberOfErrors] = TABLENOTFOUNDHMTX;
		tt->numberOfErrors++;
		tt->level1 = 0;
		return 0;
	}

	ptr=buf+tptr->offset;

	if(tptr->offset + tptr->size > tt->filelen)	{
		tt->errors[tt->numberOfErrors] = ERRORINTABLEHMTX;
		tt->numberOfErrors++;
		tt->level1 = 0;
		return 0;
	}  

	for(i=0; i<(tptr->size+3)>>2; i++) {
		
		checksum += (*ptr++)<<24;
		checksum += (*ptr++)<<16;
		checksum += (*ptr++)<<8;
		checksum += (*ptr++);
		
	}

	if(checksum != tptr->checksum) {
		tt->errors[tt->numberOfErrors] = CHECKSUMERRORHMTX;
		tt->numberOfErrors++;
	}


	if((unsigned int)(tt->numGlyphs*2+tt->numberOfHMetrics*2) != tptr->size) {
		tt->errors[tt->numberOfErrors] = ERRORINTABLEHMTX;
		tt->numberOfErrors++;
		tt->level1 = 0;
	}


	return 0;

}

int verifylocatable(unsigned char *buf, truetype_type *tt)
{
	table_type *tptr = gettable(tt,"loca");
	unsigned int i;
	unsigned char *ptr;
	unsigned long checksum = 0;

	if(!tptr && tt->version == 0x00010000) {
		tt->errors[tt->numberOfErrors] = TABLENOTFOUNDLOCA;
		tt->numberOfErrors++;
		tt->level1 = 0;
		return -1;
	}

	ptr=buf+tptr->offset;

	if(tptr->offset + tptr->size > tt->filelen)	{
		tt->errors[tt->numberOfErrors] = ERRORINTABLELOCA;
		tt->numberOfErrors++;
		tt->level1 = 0;
		return -1;
	}  

	for(i=0; i<(tptr->size+3)>>2; i++) {
		
		checksum += (*ptr++)<<24;
		checksum += (*ptr++)<<16;
		checksum += (*ptr++)<<8;
		checksum += (*ptr++);
		
	}

	if(checksum != tptr->checksum) {
		tt->errors[tt->numberOfErrors] = CHECKSUMERRORLOCA;
		tt->numberOfErrors++;
	}

	if(tt->indexToLocFormat != 0 && tt->indexToLocFormat != 1) {
		tt->errors[tt->numberOfErrors] = ERRORINTABLELOCA;
		tt->numberOfErrors++;
		tt->level1 = 0;
		return -1;
	}

	if(tt->indexToLocFormat == 0 &&
		(unsigned int)((tt->numGlyphs+1)*2) != tptr->size) {
		tt->errors[tt->numberOfErrors] = ERRORINTABLELOCA;
		tt->numberOfErrors++;
		tt->level1 = 0;
		return -1;
	}

	else if(tt->indexToLocFormat == 1 &&
		(unsigned int)(tt->numGlyphs+1)*4 != tptr->size) {
		tt->errors[tt->numberOfErrors] = ERRORINTABLELOCA;
		tt->numberOfErrors++;
		tt->level1 = 0;
		return -1;
	}


	return 0;

}

int verifyposttable(unsigned char *buf, truetype_type *tt)
{
	table_type *tptr = gettable(tt,"post");
	unsigned int i;
	unsigned char *ptr;
	unsigned long checksum = 0;
	unsigned int version;
	unsigned int nchars;
	unsigned char *endptr;
	int index;
	int maxindex;

	if(!tptr) {
		tt->errors[tt->numberOfErrors] = TABLENOTFOUNDPOST;
		tt->numberOfErrors++;
		tt->level1 = 0;
		return 0;
	}

	ptr=buf+tptr->offset;

	if(tptr->offset + tptr->size > tt->filelen)	{
		tt->errors[tt->numberOfErrors] = ERRORINTABLEPOST;
		tt->numberOfErrors++;
		tt->level1 = 0;
		return 0;
	}  

	for(i=0; i<(tptr->size+3)>>2; i++) {
		
		checksum += (*ptr++)<<24;
		checksum += (*ptr++)<<16;
		checksum += (*ptr++)<<8;
		checksum += (*ptr++);
		
	}

	if(checksum != tptr->checksum) {
		tt->errors[tt->numberOfErrors] = CHECKSUMERRORPOST;
		tt->numberOfErrors++;
	}

	ptr = buf + tptr->offset;

	if(ptr+4 > buf+tptr->offset+tptr->size){
		tt->errors[tt->numberOfErrors] = ERRORINTABLEPOST;
		tt->numberOfErrors++;
		tt->level1 = 0;
		return 0;
	}

	version = (*ptr++)<<24;
	version += (*ptr++)<<16;
	version += (*ptr++)<<8;
	version += (*ptr++);
	

	if(version == 0x10000 || version == 0x30000 || version == 0x25000)
	  return 0;

  
	if(version != 0x20000) {
		tt->errors[tt->numberOfErrors] = ERRORINTABLEPOST;
		tt->numberOfErrors++;
		tt->level1 = 0;
		return 0;
	}

	
	ptr += 28;

	if(ptr+2 > buf+tptr->offset+tptr->size){
		tt->errors[tt->numberOfErrors] = ERRORINTABLEPOST;
		tt->numberOfErrors++;
		tt->level1 = 0;
		return 0;
	}


	nchars = (*ptr++)<<8;
	nchars += (*ptr++);
	  

	if(nchars != (unsigned int)tt->numGlyphs) {
		tt->errors[tt->numberOfErrors] = ERRORINTABLEPOST;
		tt->numberOfErrors++;
		tt->level1 = 0;
		return 0;
	}

	ptr += nchars*2;
	endptr = buf + tptr->offset + tptr->size;
	maxindex = 0;

	if(endptr > buf+tptr->offset+tptr->size){
		tt->errors[tt->numberOfErrors] = ERRORINTABLEPOST;
		tt->numberOfErrors++;
		tt->level1 = 0;
		return 0;
	}

	while(ptr < endptr) {
		int stringlen = *ptr++;
		ptr += stringlen;
		maxindex++;
	}

	ptr = buf + tptr->offset + 34;
	if(ptr+2*nchars > buf+tptr->offset+tptr->size){
		tt->errors[tt->numberOfErrors] = ERRORINTABLEPOST;
		tt->numberOfErrors++;
		tt->level1 = 0;
		return 0;
	}
	for(i=0; i<nchars; i++) {
		index = (*ptr++)<<8;
		index += (*ptr++);

		if(index < 258)
			continue;
		index -= 258;
		if(index > maxindex) {
			tt->errors[tt->numberOfErrors] = ERRORINTABLEPOST;
			tt->numberOfErrors++;
			tt->level1 = 0;
			return 0;
		}
	}


    

	return 0;
}

int verifycompositeglyph(char *buf, int size)
{
	unsigned char *ptr = buf+10;
	unsigned char *endptr = buf + size;
	unsigned int i;
	unsigned short flag;

	i=0;
    do {
		if(ptr+2 > endptr)
			return -1;
		flag = (*ptr++)<<8;
		flag += (*ptr++);
		ptr += 2;  
		
		if(flag & BIT0) {
	        ptr += 4;
			
		}
		else {
	        ptr += 2;
		    
		}
		if(flag & BIT3) {
			ptr += 2;
			
		}
		else if(flag & BIT6) {
			ptr += 4;
			
		}	
		else if(flag & BIT7) {
			ptr += 8;
			
		}
		if(ptr > endptr)
			return -1;
		i++;
	} while (flag & BIT5);

    

	if(flag & BIT8) {
		int len;
		if(ptr+2 > endptr)
			return -1;
		len = (*ptr++)<<8;
		len += (*ptr++);
		ptr += len;
		if(ptr > endptr)
			return -1;
		
	}


	return 0;
}  

int verifysimpleglyph(char *buf, int size)
{
  
	unsigned char *ptr = buf;
	unsigned int i;
	unsigned int numpoints;
	unsigned int numContours;
	unsigned int len;
	unsigned char *flagptr;
	unsigned char *fptr;
	unsigned int flags;
	unsigned char *endptr = buf + size;
	unsigned int rep;


	if(ptr+4 > endptr)
		return -1;
	numContours = (*ptr++)<<8;
	numContours += (*ptr++);
  
	if(ptr + 8 + 2 * numContours > endptr)
		return -1;
	ptr += 8 + 2 * (numContours-1);

	
	numpoints = (*ptr++)<<8;
	numpoints += (*ptr++);
	numpoints++;

	if(ptr+2 > endptr)
		return -1;

	len = (*ptr++)<<8;
	len += (*ptr++);
    
	if(ptr+len > endptr)
		return -1;

    ptr += len;

	
	flagptr = ptr;
	
    
    for(i=0; i<numpoints; i++) {
		if(ptr+1 > endptr)
			return -1;
		flags = *ptr++;
		
      
		if(flags & BIT3) {
			if(ptr+1 > endptr)
				return -1;
			rep = *ptr++;
			i += rep;
			
		}
   
    } 

	fptr = flagptr;
	rep = 0;
    for(i=0; i<numpoints; i++) {
		
		if(rep == 0) {
			
			flags = *fptr++;
        }
		if(rep == 0 && flags & BIT3) {
			
			if(ptr+1 > endptr)
				return -1;
			rep = *fptr++;
			
		}
		else if(rep != 0)
			rep--;

		if(flags & BIT1) {
			if(ptr+1 > endptr)
				return -1;
			ptr++;
		}
        
        else if((flags & BIT4) == 0) {
		  
          ptr += 2;
		  if(ptr > endptr)
			return -1;
		  
        }
      
    }
	fptr = flagptr;
	rep = 0;
    for(i=0; i<numpoints; i++) {
		if(rep == 0) {
			
			flags = *fptr++;
        }
		if(rep == 0 && flags & BIT3) {
			
			if(fptr+1 > endptr)
				return -1;
			rep = *fptr++;
			
		}
		else if(rep != 0)
			rep--;

		if(flags & BIT2) {
			if(ptr+1 > endptr)
				return -1;
			ptr++;
		}
        
        else if((flags & BIT5) == 0) {
		  
          ptr += 2;
		  if(ptr > endptr)
			return -1;
		  
        }
    }
  
	return 0;
}  

int verifyglyftable(unsigned char *buf, truetype_type *tt)
{
	table_type *tptr = gettable(tt,"glyf");
	table_type *locaptr = gettable(tt,"loca");
	unsigned int i;
	unsigned char *ptr;
	unsigned long checksum = 0;
	unsigned char *lptr;
	unsigned char *gptr;
	int startoffset;
	int endoffset;
	int size;
	int numContours;
	

	if(!tptr && tt->version == 0x00010000) {
		tt->errors[tt->numberOfErrors] = TABLENOTFOUNDGLYF;
		tt->numberOfErrors++;
		tt->level1 = 0;
		return -1;
	}
	else if(!tptr)
		return 0;

	ptr=buf+tptr->offset;

	if(tptr->offset + tptr->size > tt->filelen)	{
		tt->errors[tt->numberOfErrors] = ERRORINTABLEGLYF;
		tt->numberOfErrors++;
		tt->level1 = 0;
		return 0;
	}  

	for(i=0; i<(tptr->size+3)>>2; i++) {
		
		checksum += (*ptr++)<<24;
		checksum += (*ptr++)<<16;
		checksum += (*ptr++)<<8;
		checksum += (*ptr++);
		
	}

	if(checksum != tptr->checksum) {
		tt->errors[tt->numberOfErrors] = CHECKSUMERRORGLYF;
		tt->numberOfErrors++;
	}

	ptr = buf + tptr->offset;
	lptr = buf + locaptr->offset;

	if(tt->indexToLocFormat == 0) {
		startoffset = (*lptr++)<<8;
		startoffset += (*lptr++);
		startoffset *= 2;
	}
	else {
		startoffset = (*lptr++)<<24;
		startoffset += (*lptr++)<<16;
		startoffset += (*lptr++)<<8;
		startoffset += (*lptr++);
		
	}


	for(i=0; i<(unsigned int)tt->numGlyphs; i++) {
	
		if(tt->indexToLocFormat == 0) {
			endoffset = (*lptr++)<<8;
			endoffset += (*lptr++);
			endoffset *= 2;
		}
		else {
			endoffset = (*lptr++)<<24;
			endoffset += (*lptr++)<<16;
			endoffset += (*lptr++)<<8;
			endoffset += (*lptr++);
		
		} 
	 
		size = endoffset - startoffset;

		if(size == 0) {
			startoffset = endoffset;
			
			continue;
		} 

		gptr = buf + tptr->offset + startoffset;

		if(gptr+2 > buf+tptr->offset+tptr->size)
			return -1;
	  
		numContours = (*gptr)<<8;
		numContours += (*(gptr+1));

		

		if(numContours == 0xffff) {
			if(verifycompositeglyph(gptr, size)) {
				tt->errors[tt->numberOfErrors] = ERRORINTABLEGLYF;
				tt->numberOfErrors++;
				tt->level1 = 0;
				return -1;
			}
		}
		else if(numContours) {
			if(verifysimpleglyph(gptr, size)) {
				tt->errors[tt->numberOfErrors] = ERRORINTABLEGLYF;
				tt->numberOfErrors++;
				tt->level1 = 0;
				return -1;
			}
		}
	 
		
		startoffset = endoffset;
		

	} 

	return 0;

}

int verifyebsctable(unsigned char *buf, truetype_type *tt)
{
	table_type *tptr = gettable(tt,"ebsc");
	unsigned int i;
	unsigned char *ptr;
	unsigned long checksum = 0;
	unsigned int numsizes;
	
	if(!tptr) {
		return 0;
	}

	ptr =buf+tptr->offset;

	if(tptr->offset + tptr->size > tt->filelen)	{
		tt->errors[tt->numberOfErrors] = ERRORINTABLEEBSC;
		tt->numberOfErrors++;
		return 0;
	}  

	for(i=0; i<(tptr->size+3)>>2; i++) {
		
		checksum += (*ptr++)<<24;
		checksum += (*ptr++)<<16;
		checksum += (*ptr++)<<8;
		checksum += (*ptr++);
		
	}

	if(checksum != tptr->checksum) {
		tt->errors[tt->numberOfErrors] = CHECKSUMERROREBSC;
		tt->numberOfErrors++;
	}

	ptr = buf + tptr->offset + 4;
	if(ptr+4 > buf+tptr->offset+tptr->size){
		tt->errors[tt->numberOfErrors] = ERRORINTABLEEBSC;
		tt->numberOfErrors++;
		return 0;
	}
	
	numsizes = (*ptr++)<<24;
	numsizes += (*ptr++)<<16;
	numsizes += (*ptr++)<<8;
	numsizes += (*ptr++);

	if(16 * numsizes + 8 != tptr->size) {
		tt->errors[tt->numberOfErrors] = ERRORINTABLEEBSC;
		tt->numberOfErrors++;
	}

  
	
	return 0;
}
int verifyeblcebdttable(unsigned char *buf, truetype_type *tt)
{
	table_type *tptr = gettable(tt,"EBLC");
	table_type *dtptr = gettable(tt,"EBDT");
	unsigned int i;
	unsigned int j;
	unsigned int k;
	unsigned char *ptr;
	unsigned long checksum = 0;
	unsigned int ebdtsize;
	unsigned int numsizes;
	unsigned int indexSubTableArrayOffset;
	unsigned int indexTableSize;
	unsigned int numberOfIndexSubTables;
	unsigned char *istptr;
	unsigned char *stptr;
	unsigned int firstGlyphIndex;
	unsigned int lastGlyphIndex;
	unsigned int additionalOffset;
	unsigned int indexFormat;
	unsigned int imageDataOffset;
	unsigned int offset;
	unsigned int imageSize;
	unsigned int numGlyphs;


	if(!tptr) {
		return 0;
	}
	if(!dtptr) {
		return 0;
	}
	
	ptr=buf+tptr->offset;
	ebdtsize = dtptr->size;

	if(tptr->offset + tptr->size > tt->filelen)	{
		tt->errors[tt->numberOfErrors] = ERRORINTABLEEBLC;
		tt->numberOfErrors++;
		return 0;
	}  

	for(i=0; i<(tptr->size+3)>>2; i++) {
		
		checksum += (*ptr++)<<24;
		checksum += (*ptr++)<<16;
		checksum += (*ptr++)<<8;
		checksum += (*ptr++);
		
	}

	if(checksum != tptr->checksum) {
		tt->errors[tt->numberOfErrors] = CHECKSUMERROREBLC;
		tt->numberOfErrors++;
	}

	checksum = 0;
	ptr = buf + dtptr->offset;

	for(i=0; i<(dtptr->size+3)>>2; i++) {
		
		checksum += (*ptr++)<<24;
		checksum += (*ptr++)<<16;
		checksum += (*ptr++)<<8;
		checksum += (*ptr++);
		
	}

	if(checksum != dtptr->checksum) {
		tt->errors[tt->numberOfErrors] = CHECKSUMERROREBDT;
		tt->numberOfErrors++;
	}


	ptr = buf + tptr->offset + 4;
	if(ptr+4 > buf+tptr->offset+tptr->size){
		tt->errors[tt->numberOfErrors] = ERRORINTABLEEBLC;
		tt->numberOfErrors++;
		return 0;
	}

	numsizes = (*ptr++)<<24;
	numsizes += (*ptr++)<<16;
	numsizes += (*ptr++)<<8;
	numsizes += (*ptr++);
 
	for(i=0; i<numsizes; i++) {
		if(ptr+48 > buf+tptr->offset+tptr->size){
			tt->errors[tt->numberOfErrors] = ERRORINTABLEEBLC;
			tt->numberOfErrors++;
			return 0;
		} 
		indexSubTableArrayOffset = (*ptr++)<<24;
		indexSubTableArrayOffset += (*ptr++)<<16;
		indexSubTableArrayOffset += (*ptr++)<<8;
		indexSubTableArrayOffset += (*ptr++);

		
		indexTableSize = (*ptr++)<<24;
		indexTableSize += (*ptr++)<<16;
		indexTableSize += (*ptr++)<<8;
		indexTableSize += (*ptr++);

		numberOfIndexSubTables = (*ptr++)<<24;
		numberOfIndexSubTables += (*ptr++)<<16;
		numberOfIndexSubTables += (*ptr++)<<8;
		numberOfIndexSubTables += (*ptr++);

		ptr += 36;

		

		istptr = buf + tptr->offset + indexSubTableArrayOffset;

		for(j=0; j<numberOfIndexSubTables; j++) {
			if(istptr+8 > buf+tptr->offset+tptr->size){
				tt->errors[tt->numberOfErrors] = ERRORINTABLEEBLC;
				tt->numberOfErrors++;
				return 0;
			} 
			
			firstGlyphIndex = (*istptr++)<<8;
			firstGlyphIndex += (*istptr++);

			lastGlyphIndex = (*istptr++)<<8;
			lastGlyphIndex += (*istptr++);
			
			additionalOffset = (*istptr++)<<24;
			additionalOffset += (*istptr++)<<16;
			additionalOffset += (*istptr++)<<8;
			additionalOffset += (*istptr++);

			stptr = buf + tptr->offset + indexSubTableArrayOffset + additionalOffset;

			if(stptr+8 > buf+tptr->offset+tptr->size){
				tt->errors[tt->numberOfErrors] = ERRORINTABLEEBLC;
				tt->numberOfErrors++;
				return 0;
			} 

			indexFormat = (*stptr++)<<8;
			indexFormat += (*stptr++);

			stptr += 2;

			imageDataOffset = (*stptr++)<<24;
			imageDataOffset += (*stptr++)<<16;
			imageDataOffset += (*stptr++)<<8;
			imageDataOffset += (*stptr++);
			
		
			if(indexFormat == 1) {
				unsigned int nchars = lastGlyphIndex - firstGlyphIndex +1;
			
			
				for(k=0; k<nchars; k++) {
					if(stptr+4 > buf+tptr->offset+tptr->size){
						tt->errors[tt->numberOfErrors] = ERRORINTABLEEBLC;
						tt->numberOfErrors++;
						return 0;
					} 
					offset = (*stptr++)<<24;
					offset += (*stptr++)<<16;
					offset += (*stptr++)<<8;
					offset += (*stptr++);

					if(imageDataOffset + offset > dtptr->size) {
						tt->errors[tt->numberOfErrors] = ERRORINTABLEEBDT;
						tt->numberOfErrors++;
						return 0;
					}  

				} 
			} 
			else if(indexFormat == 2) {
				int nchars = lastGlyphIndex - firstGlyphIndex +1;

				if(stptr+4 > buf+tptr->offset+tptr->size){
					tt->errors[tt->numberOfErrors] = ERRORINTABLEEBLC;
					tt->numberOfErrors++;
					return 0;
				} 

				imageSize = (*stptr++)<<24;
				imageSize += (*stptr++)<<16;
				imageSize += (*stptr++)<<8;
				imageSize += (*stptr++);

				stptr += 8;

				if(imageDataOffset + nchars * imageSize > dtptr->size) {
					tt->errors[tt->numberOfErrors] = ERRORINTABLEEBDT;
					tt->numberOfErrors++;
					return 0;
				}   

			
			
			
			} 
			else if(indexFormat == 3) {
				unsigned int nchars = lastGlyphIndex - firstGlyphIndex +1;
				if(stptr+2 > buf+tptr->offset+tptr->size){
					tt->errors[tt->numberOfErrors] = ERRORINTABLEEBLC;
					tt->numberOfErrors++;
					return 0;
				} 
			
			
				for(k=0; k<nchars; k++) {
					offset = (*stptr++)<<8;
					offset += (*stptr++);

					if(imageDataOffset + offset > dtptr->size) {
						tt->errors[tt->numberOfErrors] = ERRORINTABLEEBDT;
						tt->numberOfErrors++;
						return 0;
					}  

				} 
			}
			else if(indexFormat == 4) {
				if(stptr+4 > buf+tptr->offset+tptr->size){
					tt->errors[tt->numberOfErrors] = ERRORINTABLEEBLC;
					tt->numberOfErrors++;
					return 0;
				} 
				numGlyphs = (*stptr++)<<24;
				numGlyphs += (*stptr++)<<16;
				numGlyphs += (*stptr++)<<8;
				numGlyphs += (*stptr++);

				for(k=0; (unsigned long)numGlyphs; k++) {
					stptr += 2;

					if(stptr+2 > buf+tptr->offset+tptr->size){
						tt->errors[tt->numberOfErrors] = ERRORINTABLEEBLC;
						tt->numberOfErrors++;
						return 0;
					} 
					offset = (*stptr++)<<8;
					offset += (*stptr++);

					if(imageDataOffset + offset > dtptr->size) {
						tt->errors[tt->numberOfErrors] = ERRORINTABLEEBDT;
						tt->numberOfErrors++;
						return 0;
					}  
					
				}	 
			} 
			else if(indexFormat == 5) {


				if(stptr+16 > buf+tptr->offset+tptr->size){
					tt->errors[tt->numberOfErrors] = ERRORINTABLEEBLC;
					tt->numberOfErrors++;
					return 0;
				}

				imageSize = (*stptr++)<<24;
				imageSize += (*stptr++)<<16;
				imageSize += (*stptr++)<<8;
				imageSize += (*stptr++);

				stptr += 8;

				numGlyphs = (*stptr++)<<24;
				numGlyphs += (*stptr++)<<16;
				numGlyphs += (*stptr++)<<8;
				numGlyphs += (*stptr++);

				if(stptr+2*numGlyphs > buf+tptr->offset+tptr->size){
					tt->errors[tt->numberOfErrors] = ERRORINTABLEEBLC;
					tt->numberOfErrors++;
					return 0;
				} 
				
				if(imageDataOffset + numGlyphs*imageSize > dtptr->size) {
					tt->errors[tt->numberOfErrors] = ERRORINTABLEEBDT;
					tt->numberOfErrors++;
					return 0;
				}  
				
			} 
			

	} 

  }
  return 0;
}

int verifydevicetable(unsigned int tableoffset,unsigned char *buf, table_type *tptr)
{
	unsigned char *base=buf+tptr->offset+tableoffset;
	unsigned int startsize;
	unsigned int endsize;
	unsigned int deltaformat;
	unsigned int num;

	if(base+6 > buf+tptr->offset+tptr->size){
		return -1;
	}

	startsize = (*base++)<<8;
	startsize += (*base++);

	endsize = (*base++)<<8;
	endsize += (*base++);

	deltaformat = (*base++)<<8;
	deltaformat += (*base++);

	num = endsize - startsize + 1;

	if(deltaformat == 1) {
		num = (num + 7)>>3;
		if(tableoffset + 6 + num > tptr->offset + tptr->size)
			return -1;
	}

	else if(deltaformat == 2) {
		num = (num + 3)>>2;
		if(tableoffset + 6 + num > tptr->offset + tptr->size)
			return -1;
	}

	else if(deltaformat == 3) {
		num = (num + 1)>>1;
		if(tableoffset + 6 + num > tptr->offset + tptr->size)
			return -1;
	}

	else 
		return -1;

	return 0;
}


int verifyclassdeftable(unsigned int tableoffset,unsigned char *buf, table_type *tptr)
{
	unsigned char *base=buf+tptr->offset+tableoffset;
	unsigned char *ptr = base;
	unsigned int count;
	unsigned int format;

	if(ptr+2 > buf+tptr->offset+tptr->size){

		return -1;
	}

	format = (*ptr++)<<8;
	format += (*ptr++);

	if(format == 1) {
		ptr += 2;

		if(ptr+2 > buf+tptr->offset+tptr->size){
			return -1;
		} 

		count = (*ptr++)<<8;
		count += (*ptr++);

		if(ptr + count * 2 > buf + tptr->offset+tptr->size)
			return -1;
	}
	else if(format == 2) {
		if(ptr+2 > buf+tptr->offset+tptr->size){
			return -1;
		} 
		count = (*ptr++)<<8;
		count += (*ptr++);

		if(ptr + count * 6 > buf + tptr->offset+tptr->size)
			return -1;
	}

	else
		return -1;


	return 0;
}

int verifycoveragetable(unsigned int tableoffset,unsigned char *buf, table_type *tptr)
{
	unsigned char *base=buf+tptr->offset+tableoffset;
	unsigned char *ptr = base;
	unsigned int count;
	unsigned int format;

	if(ptr+2 > buf+tptr->offset+tptr->size){
		return -1;
	}

	format = (*ptr++)<<8;
	format += (*ptr++);

	if(format == 1) {
		
		if(ptr+2 > buf+tptr->offset+tptr->size){
			return -1;
		} 

		count = (*ptr++)<<8;
		count += (*ptr++);

		if(ptr + count * 2 > buf + tptr->offset+tptr->size)
			return -1;
	}
	else if(format == 2) {

		if(ptr+2 > buf+tptr->offset+tptr->size){
			return -1;
		} 
		count = (*ptr++)<<8;
		count += (*ptr++);

		if(ptr + count * 6 > buf + tptr->offset+tptr->size)
			return -1;
	}

	else
		return -1;



	return 0;
}


int verifyligcaretlisttable(unsigned int tableoffset,unsigned char *buf, table_type *tptr)
{
	unsigned char *base=buf+tptr->offset+tableoffset;
	unsigned char *ptr = base;
	unsigned int offset;
	unsigned int count;
	unsigned int i;
	unsigned int j;
	unsigned char *ligptr;
	unsigned int cvoffset;
	unsigned char *cvptr;
	unsigned int caretformat;
	unsigned int deviceoffset;

	if(ptr+4 > buf+tptr->offset+tptr->size){
		return -1;
	}

	offset = (*ptr++)<<8;
	offset += (*ptr++);

	if(offset) {
		if(verifycoveragetable(tableoffset+offset,buf,tptr))
			return -1;
	}


	count = (*ptr++)<<8;
	count += (*ptr++);

	for(i=0; i<count; i++) {
		if(ptr+2 > buf+tptr->offset+tptr->size){
			return -1;
		} 

		offset = (*ptr++)<<8;
		offset += (*ptr++);

		if(!offset)
			continue;

		ligptr = base + offset;

		if(ligptr+2 > buf+tptr->offset+tptr->size){
			return -1;
		} 

		count = (*ligptr++)<<8;
		count += (*ligptr++);

		for(j=0; j<count; j++) {
			if(ligptr+2 > buf+tptr->offset+tptr->size){
				return -1;
			} 
			cvoffset = (*ligptr++)<<8;
			cvoffset += (*ligptr++);

			if(!cvoffset)
				continue;

			cvptr = base + offset + cvoffset;
			if(cvptr+2 > buf+tptr->offset+tptr->size){
				return -1;
			} 

			caretformat = (*cvptr++)<<8;
			caretformat += (*cvptr++);

			if(caretformat == 1 || caretformat == 2) {
				if(cvptr + 2 > buf+tptr->offset+tptr->size)
					return -1;
			}

			else if(caretformat == 3) {
				cvptr += 2;
				if(cvptr+2 > buf+tptr->offset+tptr->size){
					return -1;
				} 
				
				deviceoffset = (*cvptr++)<<8;
				deviceoffset += (*cvptr++);

				if(deviceoffset) {
					if(verifydevicetable(deviceoffset + cvoffset + offset,buf,tptr))
						return -1;
				}
			}
			else
				return -1;
		}
	}

	return 0;
}

verifyattachlisttable(unsigned int tableoffset,unsigned char *buf, table_type *tptr)
{
	unsigned char *base=buf+tptr->offset+tableoffset;
	unsigned char *ptr = base;
	unsigned int offset;
	unsigned int count;
	unsigned int i;
	unsigned char *pointptr;

	if(ptr+4 > buf+tptr->offset+tptr->size){
		return -1;
	}

	offset = (*ptr++)<<8;
	offset += (*ptr++);

	if(offset) {
		if(verifycoveragetable(tableoffset+offset,buf,tptr))
			return -1;
	}


	count = (*ptr++)<<8;
	count += (*ptr++);

	for(i=0; i<count; i++) {

		if(ptr+2 > buf+tptr->offset+tptr->size){
			return -1;
		} 

		offset = (*ptr++)<<8;
		offset += (*ptr++);

		if(!offset)
			continue;

		pointptr = base + offset;

		if(pointptr+2 > buf+tptr->offset+tptr->size){
			return -1;
		} 

		count = (*pointptr++)<<8;
		count += (*pointptr++);

		if(pointptr + count * 2 > buf+tptr->offset + tptr->size)
			return -1;
	}

	return 0;
}




int verifygdeftable(unsigned char *buf, truetype_type *tt)
{
	table_type *tptr = gettable(tt,"GDEF");
	unsigned int i;
	unsigned char *ptr;
	unsigned long checksum = 0;
	unsigned int offset;
	
	if(!tptr) {
		return 0;
	}

	ptr=buf+tptr->offset;

	if(tptr->offset + tptr->size > tt->filelen)	{
		tt->errors[tt->numberOfErrors] = ERRORINTABLEGDEF;
		tt->numberOfErrors++;
		return 0;
	}  

	for(i=0; i<(tptr->size+3)>>2; i++) {
		
		checksum += (*ptr++)<<24;
		checksum += (*ptr++)<<16;
		checksum += (*ptr++)<<8;
		checksum += (*ptr++);
		
	}

	if(checksum != tptr->checksum) {
		tt->errors[tt->numberOfErrors] = CHECKSUMERRORGDEF;
		tt->numberOfErrors++;
	}

	ptr = buf + tptr->offset + 4;

	if(ptr+8 > buf+tptr->offset+tptr->size){
		tt->errors[tt->numberOfErrors] = ERRORINTABLEGDEF;
		tt->numberOfErrors++;
		return 0;
	}

	offset = (*ptr++)<<8;
	offset += (*ptr++);

	if(offset) {
		if(verifyclassdeftable(offset,buf,tptr)) {
			tt->errors[tt->numberOfErrors] = ERRORINTABLEGDEF;
			tt->numberOfErrors++;
			return 0;
		} 
	}

	offset = (*ptr++)<<8;
	offset += (*ptr++);

	if(offset) {
		if(verifyattachlisttable(offset,buf,tptr)) {
			tt->errors[tt->numberOfErrors] = ERRORINTABLEGDEF;
			tt->numberOfErrors++;
			return 0;
		} 
	}

	offset = (*ptr++)<<8;
	offset += (*ptr++);

	if(offset) {
		if(verifyligcaretlisttable(offset,buf,tptr)) {
			tt->errors[tt->numberOfErrors] = ERRORINTABLEGDEF;
			tt->numberOfErrors++;
			return 0;
		} 
	}

	offset = (*ptr++)<<8;
	offset += (*ptr++);

	if(offset) {
		if(verifyclassdeftable(offset,buf,tptr)) {
			tt->errors[tt->numberOfErrors] = ERRORINTABLEGDEF;
			tt->numberOfErrors++;
			return 0;
		} 
	}

	return 0;
}


int verifyjstflangsystable(char *ptr,char *buf,table_type *tptr)
{

	unsigned int offset;
	unsigned int count;
	unsigned char *modptr;

	if(ptr+20 > buf+tptr->offset+tptr->size){
		return -1;
	}

	/* ShrinkageEnableGSUB */
	offset = (*ptr++)<<8;
	offset += (*ptr++);

	if(offset) {

		modptr = ptr + offset;

		if(modptr+2 > buf+tptr->offset+tptr->size){
			return -1;
		} 

		count = (*modptr++)<<8;
		count += (*modptr++);

		if(modptr + count * 2 > buf + tptr->offset + tptr->size)
			return -1;
	}

	/* ShrinkageDisableGSUB */
	offset = (*ptr++)<<8;
	offset += (*ptr++);

	if(offset) {

		modptr = ptr + offset;
		if(modptr+2 > buf+tptr->offset+tptr->size){
			return -1;
		} 

		count = (*modptr++)<<8;
		count += (*modptr++);

		if(modptr + count * 2 > buf + tptr->offset + tptr->size)
			return -1;
	}
	/* ShrinkageEnableGPOS */
	offset = (*ptr++)<<8;
	offset += (*ptr++);

	if(offset) {

		modptr = ptr + offset;
		if(modptr+2 > buf+tptr->offset+tptr->size){
			return -1;
		} 

		count = (*modptr++)<<8;
		count += (*modptr++);

		if(modptr + count * 2 > buf + tptr->offset + tptr->size)
			return -1;
	}
	/* ShrinkageDisableGPOS */
	offset = (*ptr++)<<8;
	offset += (*ptr++);

	if(offset) {

		modptr = ptr + offset;
		if(modptr+2 > buf+tptr->offset+tptr->size){
			return -1;
		} 

		count = (*modptr++)<<8;
		count += (*modptr++);

		if(modptr + count * 2 > buf + tptr->offset + tptr->size)
			return -1;
	}
	/* ShrinkageJstfMax */
	offset = (*ptr++)<<8;
	offset += (*ptr++);

	if(offset) {

		modptr = ptr + offset;
		if(modptr+2 > buf+tptr->offset+tptr->size){
			return -1;
		} 

		count = (*modptr++)<<8;
		count += (*modptr++);

		if(modptr + count * 2 > buf + tptr->offset + tptr->size)
			return -1;
	}

	/* ExtensionEnableGSUB */
	offset = (*ptr++)<<8;
	offset += (*ptr++);

	if(offset) {

		modptr = ptr + offset;
		if(modptr+2 > buf+tptr->offset+tptr->size){
			return -1;
		} 

		count = (*modptr++)<<8;
		count += (*modptr++);

		if(modptr + count * 2 > buf + tptr->offset + tptr->size)
			return -1;
	}

	/* ExtensionDisableGSUB */
	offset = (*ptr++)<<8;
	offset += (*ptr++);

	if(offset) {

		modptr = ptr + offset;
		if(modptr+2 > buf+tptr->offset+tptr->size){
			return -1;
		} 

		count = (*modptr++)<<8;
		count += (*modptr++);

		if(modptr + count * 2 > buf + tptr->offset + tptr->size)
			return -1;
	}

	/* ExtensionEnableGPOS */
	offset = (*ptr++)<<8;
	offset += (*ptr++);

	if(offset) {

		modptr = ptr + offset;
		if(modptr+2 > buf+tptr->offset+tptr->size){
			return -1;
		} 

		count = (*modptr++)<<8;
		count += (*modptr++);

		if(modptr + count * 2 > buf + tptr->offset + tptr->size)
			return -1;
	}

	/* ExtensionDisableGPOS */
	offset = (*ptr++)<<8;
	offset += (*ptr++);

	if(offset) {

		modptr = ptr + offset;
		if(modptr+2 > buf+tptr->offset+tptr->size){
			return -1;
		} 

		count = (*modptr++)<<8;
		count += (*modptr++);

		if(modptr + count * 2 > buf + tptr->offset + tptr->size)
			return -1;
	}

	/* ExtensionJsftMax */
	offset = (*ptr++)<<8;
	offset += (*ptr++);

	if(offset) {

		modptr = ptr + offset;
		if(modptr+2 > buf+tptr->offset+tptr->size){
			return -1;
		} 

		count = (*modptr++)<<8;
		count += (*modptr++);

		if(modptr + count * 2 > buf + tptr->offset + tptr->size)
			return -1;
	}

	return 0;
}



int verifybasecoordtable(unsigned int tableoffset,unsigned char *buf, table_type *tptr) 
{
	unsigned char *base=buf+tptr->offset+tableoffset;
	
	unsigned char *bcptr;
	unsigned int basecoordformat;
	unsigned int devicetaboffset;

	bcptr = base;
	if(bcptr+2 > buf+tptr->offset+tptr->size){
		return -1;
	}
	basecoordformat = (*bcptr++)<<8;
	basecoordformat += (*bcptr++);
	if(basecoordformat == 1) {
		if(bcptr + 2 > buf + tptr->offset + tptr->size)
			return -1;
	}
	else if(basecoordformat == 2) {
		if(bcptr + 6 > buf + tptr->offset + tptr->size)
			return -1;
	}
	else if(basecoordformat == 3) {
		bcptr += 2;

		if(bcptr+2 > buf+tptr->offset+tptr->size){
			return -1;
		} 
		devicetaboffset = (*bcptr++)<<8;
		devicetaboffset += (*bcptr++);

		if(devicetaboffset) {
			if(verifydevicetable(devicetaboffset + tableoffset + 
				tptr->offset,buf,tptr))
				return -1;
		}

						
	}

	return 0;
}


int verifyminmaxtable(unsigned int tableoffset,unsigned char *buf, table_type *tptr) 
{
	unsigned char *base=buf+tptr->offset+tableoffset;
	unsigned int offset;
	unsigned int count;
	unsigned int i;

	if(base+6 > buf+tptr->offset+tptr->size){
		return -1;
	}

	/* MinCoord */
	offset = (*base++)<<8;
	offset += (*base++);

	if(offset) {
		if(verifybasecoordtable(tableoffset + offset,buf,tptr))
			return -1;
	}

	/* MaxCoord */
	offset = (*base++)<<8;
	offset += (*base++);

	if(offset) {
		if(verifybasecoordtable(tableoffset + offset,buf,tptr))
			return -1;
	}

	count = (*base++)<<8;
	count += (*base++);

	for(i=0; i<count; i++) {
		base += 2;

		if(base+4 > buf+tptr->offset+tptr->size){
			return -1;
		} 

		/* MinCoord */
		offset = (*base++)<<8;
		offset += (*base++);

		if(offset) {
			if(verifybasecoordtable(tableoffset + offset,buf,tptr))
				return -1;
		}

		/* MaxCoord */
		offset = (*base++)<<8;
		offset += (*base++);

		if(offset) {
			if(verifybasecoordtable(tableoffset + offset,buf,tptr))
				return -1;
		}
	}

	return 0;
}




int verifyaxistable(unsigned int tableoffset,unsigned char *buf, table_type *tptr) 
{
	unsigned int i;
	unsigned char *base=buf+tptr->offset+tableoffset;
	unsigned char *ptr = base;
	unsigned int offset;
	unsigned int count;
	unsigned char *bptr;
	unsigned int basescriptoffset;
	unsigned char *basescriptptr;
	unsigned int basevaluesoffset;
	unsigned char *bvptr;
	unsigned int bccount;
	unsigned int j;
	unsigned int basecoordoffset;
	unsigned int baselangsyscount;
	unsigned int minmaxoffset;

	/* BaseTagList */
	if(ptr+4 > buf+tptr->offset+tptr->size){
		return -1;
	}
	offset = (*ptr++)<<8;
	offset += (*ptr++);

	if(offset) {
		bptr = base + offset;

		if(bptr+2 > buf+tptr->offset+tptr->size){
			return -1;
		} 
		
		count = (*bptr++)<<8;
		count += (*bptr++);

		if(bptr + count * 4 > buf + tptr->offset + tptr->size)
			return -1;
	}

	/* BaseScriptList */
	offset = (*ptr++)<<8;
	offset += (*ptr++);

	if(offset) {
		bptr = base + offset;

		if(bptr+2 > buf+tptr->offset+tptr->size){
			return -1;
		} 
		
		count = (*bptr++)<<8;
		count += (*bptr++);

		for(i=0; i<count; i++) {
			bptr += 4;

			if(bptr+2> buf+tptr->offset+tptr->size){
				return -1;
			} 

			basescriptoffset = (*bptr++)<<8;
			basescriptoffset += (*bptr++);

			if(!basescriptoffset)
				continue;

			basescriptptr = base + offset + basescriptoffset;

			if(basescriptptr+6 > buf+tptr->offset+tptr->size){
				return -1;
			} 

			basevaluesoffset = (*basescriptptr++)<<8;
			basevaluesoffset += (*basescriptptr++);

			if(basevaluesoffset) {
				bvptr = base + offset + basescriptoffset + basevaluesoffset + 2;

				if(bvptr+2 > buf+tptr->offset+tptr->size){
					return -1;
				} 
				bccount = (*bvptr++)<<8;
				bccount += (*bvptr++);

				for(j=0; j<bccount; j++) {
					if(bvptr+2 > buf+tptr->offset+tptr->size){
						return -1;
					}  
					basecoordoffset = (*bvptr++)<<8;
					basecoordoffset += (*bvptr++);

					if(basecoordoffset) {
						if(verifybasecoordtable(tableoffset+offset+basescriptoffset+basevaluesoffset,buf,tptr))
							return -1;
					}

					
				}
			}


			/* default MinMax */
			minmaxoffset = (*basescriptptr++)<<8;
			minmaxoffset += (*basescriptptr++);

			if(minmaxoffset) {
				if(verifyminmaxtable(tableoffset+offset+basescriptoffset+minmaxoffset,buf,tptr))
					return -1;
			}

			baselangsyscount = (*basescriptptr++)<<8;
			baselangsyscount += (*basescriptptr++);

			for(j=0; j<baselangsyscount; j++) {
				basescriptptr += 4;
				if(basescriptptr+2 > buf+tptr->offset+tptr->size){
					return -1;
				} 
				minmaxoffset = (*basescriptptr++)<<8;
				minmaxoffset += (*basescriptptr++);
			
				if(minmaxoffset) {
					if(verifyminmaxtable(tableoffset+offset+basescriptoffset+minmaxoffset,buf,tptr))
						return -1;
				}

			}
		}
	}


	return 0;
}

int verifybasetable(unsigned char *buf, truetype_type *tt)
{
	table_type *tptr = gettable(tt,"BASE");
	unsigned int i;
	unsigned char *ptr;
	unsigned long checksum = 0;
	unsigned int offset;
	
	if(!tptr) {
		return 0;
	}

	ptr=buf+tptr->offset;

	if(tptr->offset + tptr->size > tt->filelen)	{
		tt->errors[tt->numberOfErrors] = ERRORINTABLEBASE;
		tt->numberOfErrors++;
		return 0;
	}  

	for(i=0; i<(tptr->size+3)>>2; i++) {
		
		checksum += (*ptr++)<<24;
		checksum += (*ptr++)<<16;
		checksum += (*ptr++)<<8;
		checksum += (*ptr++);
		
	}

	if(checksum != tptr->checksum) {
		tt->errors[tt->numberOfErrors] = CHECKSUMERRORBASE;
		tt->numberOfErrors++;
	}

	ptr = buf + tptr->offset + 4;

	if(ptr+4 > buf+tptr->offset+tptr->size){
		tt->errors[tt->numberOfErrors] = ERRORINTABLEBASE;
		tt->numberOfErrors++;
		return 0;
	}

	offset = (*ptr++)<<8;
	offset += (*ptr++);

	if(offset) {
		if(verifyaxistable(offset,buf,tptr)) {
			tt->errors[tt->numberOfErrors] = ERRORINTABLEBASE;
			tt->numberOfErrors++;
			return 0;
		}
	}

	offset = (*ptr++)<<8;
	offset += (*ptr++);

	if(offset) {
		if(verifyaxistable(offset,buf,tptr)) {
			tt->errors[tt->numberOfErrors] = ERRORINTABLEBASE;
			tt->numberOfErrors++;
			return 0;
		}
	}

	return 0;
}

int verifyjstftable(unsigned char *buf, truetype_type *tt)
{
	table_type *tptr = gettable(tt,"JSTF");
	unsigned int i;
	unsigned char *ptr;
	unsigned long checksum = 0;
	unsigned int count;
	unsigned int offset;
	unsigned char *jstptr;
	unsigned char *extptr;
	unsigned int glyphcount;
	unsigned int extoffset;
	unsigned int defoffset;
	unsigned int langsyscount;
	unsigned int j;
	unsigned int langsysoffset;

	
	if(!tptr) {
		return 0;
	}

	ptr=buf+tptr->offset;

	if(tptr->offset + tptr->size > tt->filelen)	{
		tt->errors[tt->numberOfErrors] = ERRORINTABLEJSTF;
		tt->numberOfErrors++;
		return 0;
	}  

	for(i=0; i<(tptr->size+3)>>2; i++) {
		
		checksum += (*ptr++)<<24;
		checksum += (*ptr++)<<16;
		checksum += (*ptr++)<<8;
		checksum += (*ptr++);
		
	}

	if(checksum != tptr->checksum) {
		tt->errors[tt->numberOfErrors] = CHECKSUMERRORJSTF;
		tt->numberOfErrors++;
	}

	ptr = buf + tptr->offset + 4;

	if(ptr+2 > buf+tptr->offset+tptr->size){
		tt->errors[tt->numberOfErrors] = ERRORINTABLEJSTF;
		tt->numberOfErrors++;
		return 0;
	}

	count = (*ptr++)<<8;
	count += (*ptr++);

	for(i=0; i<count; i++) {
		ptr += 4;

		if(ptr+2 > buf+tptr->offset+tptr->size){
			tt->errors[tt->numberOfErrors] = ERRORINTABLEJSTF;
			tt->numberOfErrors++;
			return 0;
		} 

		offset = (*ptr++)<<8;
		offset += (*ptr++);

		if(!offset)
			continue;

		jstptr = buf + tptr->offset + offset;

		
		if(jstptr+6 > buf+tptr->offset+tptr->size){
			tt->errors[tt->numberOfErrors] = ERRORINTABLEJSTF;
			tt->numberOfErrors++;
			return 0;
		} 

		extoffset = (*jstptr++)<<8;
		extoffset += (*jstptr++);

		if(extoffset) {

			extptr = buf + tptr->offset + offset + extoffset;

			if(extptr+2 > buf+tptr->offset+tptr->size){
				tt->errors[tt->numberOfErrors] = ERRORINTABLEJSTF;
				tt->numberOfErrors++;
				return 0;
			} 
		

			glyphcount = (*extptr++)<<8;
			glyphcount += (*extptr++);

			if(extptr + 2 * glyphcount > buf + tptr->offset + tptr->size)  {
				tt->errors[tt->numberOfErrors] = ERRORINTABLEJSTF;
				tt->numberOfErrors++;
			} 
		}


		defoffset = (*jstptr++)<<8;
		defoffset += (*jstptr++);

		if(defoffset) {
			if(verifyjstflangsystable(buf + tptr->offset + offset + defoffset,buf,tptr))
				return -1;
		}


		langsyscount = (*jstptr++)<<8;
		langsyscount += (*jstptr++);

		for(j=0; j<langsyscount; j++) {
			jstptr += 4;
			if(jstptr+2 > buf+tptr->offset+tptr->size){
				tt->errors[tt->numberOfErrors] = ERRORINTABLEJSTF;
				tt->numberOfErrors++;
				return 0;
			}  
			langsysoffset = (*jstptr++)<<8;
			langsysoffset += (*jstptr++);

			if(langsysoffset) {
				if(verifyjstflangsystable(buf + tptr->offset + offset,buf,tptr)) {
					tt->errors[tt->numberOfErrors] = ERRORINTABLEJSTF;
					tt->numberOfErrors++;
				}
			}  

		}  
	}

	return 0;
}
int getvaluerecordinfo(unsigned int format, unsigned char *ptr, unsigned int *len, 
					   unsigned int *offset1, unsigned int *offset2,
					   unsigned int *offset3, unsigned int *offset4,
					   unsigned char *buf, table_type *tptr) 
{
	*len=0;
	*offset1 = 0;
	*offset2 = 0;
	*offset3 = 0;
	*offset4 = 0;

	if(format&1) {
		*len += 2;
		ptr+=2;
	}
	if(format&2) {
		*len += 2;
		ptr+=2;
	}
	if(format&4) {
		*len += 2;
		ptr+=2;
	}
	if(format&8) {
		*len += 2;
		ptr+=2;
	}
	if(format&0x10) {
		*len += 2;
		ptr+=2;
		if(ptr+2 > buf+tptr->offset+tptr->size){
			return -1;
		} 
		*offset1 = (*ptr++)<<8;
		*offset1 += (*ptr++);
		
	}
	if(format&0x20) {
		*len += 2;
		ptr+=2;
		if(ptr+2 > buf+tptr->offset+tptr->size){
			return -1;
		} 
		*offset2 = (*ptr++)<<8;
		*offset2 += (*ptr++);
		
	}
	if(format&0x40) {
		*len += 2;
		ptr+=2;
		if(ptr+2 > buf+tptr->offset+tptr->size){
			return -1;
		} 
		*offset3 = (*ptr++)<<8;
		*offset3 += (*ptr++);
		
	}
	if(format&0x80) {
		*len += 2;
		ptr+=2;
		if(ptr+2 > buf+tptr->offset+tptr->size){
			return -1;
		} 
		*offset4 = (*ptr++)<<8;
		*offset4 += (*ptr++);
		
	}

	return 0;
}
int verifyanchortable(unsigned int tableoffset,unsigned char *buf, table_type *tptr) 
{
	unsigned char *base=buf+tptr->offset+tableoffset;
	unsigned char *ptr = base;
	unsigned int format;
	unsigned int offset;

	if(ptr+2 > buf+tptr->offset+tptr->size)
		return -1;
	format = (*ptr++)<<8;
	format += (*ptr++);

	if(format == 1) {
		if(ptr+4 > buf+tptr->offset+tptr->size)
			return -1;
	}
	else if(format == 2) {
		if(ptr+6 > buf+tptr->offset+tptr->size)
			return -1;
	}
	else if(format == 3) {
		if(ptr+8 > buf+tptr->offset+tptr->size)
			return -1;
		ptr += 4;

		if(ptr+4 > buf+tptr->offset+tptr->size)
		return -1;

		offset = (*ptr++)<<8;
		offset += (*ptr++);
		if(offset) {
			if(verifydevicetable(tableoffset+offset,buf,tptr))
				return -1;
		}

		offset = (*ptr++)<<8;
		offset += (*ptr++);
		if(offset) {
			if(verifydevicetable(tableoffset+offset,buf,tptr))
				return -1;
		}
	
	}
	else
		return -1;
	return 0;
}

int verifymarkarray(unsigned int tableoffset,unsigned char *buf, table_type *tptr) 
{
	unsigned char *base=buf+tptr->offset+tableoffset;
	unsigned char *ptr = base;
	unsigned int count;
	unsigned int offset;
	unsigned int i;

	if(ptr+2 > buf+tptr->offset+tptr->size)
		return -1;

	count = (*ptr++)<<8;
	count += (*ptr++);

	for(i=0; i<count; i++) {
		if(ptr+4 > buf + tptr->offset+tptr->size)
			return -1;

		ptr += 2;

		offset = (*ptr++)<<8;
		offset += (*ptr++);

		if(offset) {
			if(verifyanchortable(tableoffset+offset,buf,tptr))
				return -1;
		}
	}
	return 0;
}



		

int verifygposlookup1(unsigned int tableoffset,unsigned char *buf, table_type *tptr) 
{
	unsigned char *base=buf+tptr->offset+tableoffset;
	unsigned char *ptr = base;
	unsigned int format;
	unsigned int coverageoffset;
	unsigned int valueformat;
	unsigned int count;
	unsigned int offset1;
	unsigned int offset2;
	unsigned int offset3;
	unsigned int offset4;
	unsigned int valuecount;
	unsigned int i;

	if(ptr+4 > buf+tptr->offset+tptr->size)
		return -1;

	format = (*ptr++)<<8;
	format += (*ptr++);

	coverageoffset = (*ptr++)<<8;
	coverageoffset += (*ptr++);

	if(coverageoffset) {
		if(verifycoveragetable(tableoffset+coverageoffset,buf,tptr))
			return -1;
	}

	if(format == 1) {
		if(ptr+2 > buf+tptr->offset+tptr->size)
			return -1;

		valueformat = (*ptr++)<<8;
		valueformat += (*ptr++);

		if(getvaluerecordinfo(valueformat,ptr,&count,&offset1,&offset2,&offset3,&offset4,buf,tptr))
			return -1;

		if(ptr+count > buf+tptr->offset+tptr->size)
			return -1;

		if(offset1) {
			if(verifydevicetable(tableoffset+offset1,buf,tptr))
				return -1;
		}

		if(offset2) {
			if(verifydevicetable(tableoffset+offset2,buf,tptr))
				return -1;
		}

		if(offset3) {
			if(verifydevicetable(tableoffset+offset3,buf,tptr))
				return -1;
		}

		if(offset4) {
			if(verifydevicetable(tableoffset+offset4,buf,tptr))
				return -1;
		}

		

	}

	else if(format == 2) {
		if(ptr+4 > buf+tptr->offset+tptr->size)
			return -1;
		valueformat = (*ptr++)<<8;
		valueformat += (*ptr++);

		valuecount = (*ptr++)<<8;
		valuecount += (*ptr++);

		for(i=0; i<valuecount; i++) {

			if(getvaluerecordinfo(valueformat,ptr,&count,&offset1,&offset2,&offset3,&offset4,buf,tptr))
				return -1;

			if(ptr+count > buf+tptr->offset+tptr->size)
				return -1;

			if(offset1) {
				if(verifydevicetable(tableoffset+offset1,buf,tptr))
					return -1;
			}

			if(offset2) {
				if(verifydevicetable(tableoffset+offset2,buf,tptr))
					return -1;
			}

			if(offset3) {
				if(verifydevicetable(tableoffset+offset3,buf,tptr))
					return -1;
			}

			if(offset4) {
				if(verifydevicetable(tableoffset+offset4,buf,tptr))
					return -1;
			}
		}
	}
	else
		return -1;


	return 0;
}

int verifygposlookup2(unsigned int tableoffset,unsigned char *buf, table_type *tptr) 
{
	unsigned char *base=buf+tptr->offset+tableoffset;
	unsigned char *ptr = base;
	unsigned int format;
	unsigned int count;
	unsigned int i;
	unsigned int j;
	unsigned int offset;
	unsigned char *psptr;
	unsigned int coverageoffset;
	unsigned int valueformat1;
	unsigned int valueformat2;
	unsigned int vrcount;
	unsigned int pvcount;
	unsigned int offset1;
	unsigned int offset2;
	unsigned int offset3;
	unsigned int offset4;
	unsigned int class1count;
	unsigned int class2count;

	if(ptr+4 > buf+tptr->offset+tptr->size)
		return -1;

	format = (*ptr++)<<8;
	format += (*ptr++);

	coverageoffset = (*ptr++)<<8;
	coverageoffset += (*ptr++);

	if(coverageoffset) {

		if(verifycoveragetable(tableoffset+coverageoffset,buf,tptr))
			return -1;
	}

	if(format == 1) {
		if(ptr+6 > buf+tptr->offset+tptr->size)
		return -1;
		valueformat1 = (*ptr++)<<8;
		valueformat1 += (*ptr++);

		valueformat2 = (*ptr++)<<8;
		valueformat2 += (*ptr++);

		count = (*ptr++)<<8;
		count += (*ptr++);

		for(i=0; i<count; i++) {

			if(ptr+2 > buf+tptr->offset+tptr->size)
				return -1;

			offset = (*ptr++)<<8;
			offset += (*ptr++);

			if(!offset)
				continue;

			psptr = base  + offset;

			if(psptr+2 > buf+tptr->offset+tptr->size)
				return -1;

			pvcount = (*psptr++)<<8;
			pvcount += (*psptr++);

			for(j=0; j<pvcount; j++) {

				psptr += 2;


				if(getvaluerecordinfo(valueformat1,psptr,&vrcount,&offset1,&offset2,&offset3,&offset4,buf,tptr))
					return -1;

				if(psptr+vrcount > buf+tptr->offset+tptr->size)
					return -1;

				if(offset1) {
					if(verifydevicetable(tableoffset+offset1,buf,tptr))
						return -1;
				} 

				if(offset2) {
					if(verifydevicetable(tableoffset+offset2,buf,tptr))
						return -1;
				}

				if(offset3) {
					if(verifydevicetable(tableoffset+offset3,buf,tptr))
						return -1;
				}

				if(offset4) {
					if(verifydevicetable(tableoffset+offset4,buf,tptr))
						return -1;
				}

				psptr += vrcount;

				
				if(getvaluerecordinfo(valueformat2,psptr,&vrcount,&offset1,&offset2,&offset3,&offset4,buf,tptr))
					return -1;

				if(psptr+vrcount > buf+tptr->offset+tptr->size)
					return -1;

				if(offset1) {
					if(verifydevicetable(tableoffset+offset1,buf,tptr))
						return -1;
				} 

				if(offset2) {
					if(verifydevicetable(tableoffset+offset2,buf,tptr))
						return -1;
				}

				if(offset3) {
					if(verifydevicetable(tableoffset+offset3,buf,tptr))
						return -1;
				}

				if(offset4) {
					if(verifydevicetable(tableoffset+offset4,buf,tptr))
						return -1;
				}
			}

		}

		

	}

	else if(format == 2) {

		if(ptr+12 > buf+tptr->offset+tptr->size)
			return -1;
		valueformat1 = (*ptr++)<<8;
		valueformat1 += (*ptr++);

		valueformat2 = (*ptr++)<<8;
		valueformat2 += (*ptr++);

		offset = (*ptr++)<<8;
		offset += (*ptr++);

		if(offset) {

			if(verifyclassdeftable(tableoffset+offset,buf,tptr))
				return -1;
		}

		offset = (*ptr++)<<8;
		offset += (*ptr++);

		if(offset) {
			if(verifyclassdeftable(tableoffset+offset,buf,tptr))
				return -1;
		}

		class1count = (*ptr++)<<8;
		class1count += (*ptr++);

		class2count = (*ptr++)<<8;
		class2count += (*ptr++);

		for(i=0; i<class1count; i++) {
			for(j=0; j<class2count; j++) {
				if(getvaluerecordinfo(valueformat1,ptr,&vrcount,&offset1,&offset2,&offset3,&offset4,buf,tptr))
					return -1;

				if(ptr+vrcount > buf+tptr->offset+tptr->size)
					return -1;

				if(offset1) {
					if(verifydevicetable(tableoffset+offset1,buf,tptr))
						return -1;
				} 

				if(offset2) {
					if(verifydevicetable(tableoffset+offset2,buf,tptr))
						return -1;
				}

				if(offset3) {
					if(verifydevicetable(tableoffset+offset3,buf,tptr))
						return -1;
				}

				if(offset4) {
					if(verifydevicetable(tableoffset+offset4,buf,tptr))
						return -1;
				}

				ptr += vrcount;

				if(getvaluerecordinfo(valueformat2,ptr,&vrcount,&offset1,&offset2,&offset3,&offset4,buf,tptr))
					return -1;

				if(ptr+vrcount > buf+tptr->offset+tptr->size)
					return -1;

				if(offset1) {
					if(verifydevicetable(tableoffset+offset1,buf,tptr))
						return -1;
				} 

				if(offset2) {
					if(verifydevicetable(tableoffset+offset2,buf,tptr))
						return -1;
				}

				if(offset3) {
					if(verifydevicetable(tableoffset+offset3,buf,tptr))
						return -1;
				}

				if(offset4) {
					if(verifydevicetable(tableoffset+offset4,buf,tptr))
						return -1;
				}

			}
		}

	}
	else
		return -1;


	return 0;
}

int verifygposlookup3(unsigned int tableoffset,unsigned char *buf, table_type *tptr) 
{
	unsigned char *base=buf+tptr->offset+tableoffset;
	unsigned char *ptr = base;
	unsigned int format;
	unsigned int count;
	unsigned int i;
	unsigned int offset;
	unsigned int coverageoffset;

	if(ptr+4 > buf+tptr->offset+tptr->size)
		return -1;

	format = (*ptr++)<<8;
	format += (*ptr++);

	coverageoffset = (*ptr++)<<8;
	coverageoffset += (*ptr++);

	if(coverageoffset) {
		if(verifycoveragetable(tableoffset+coverageoffset,buf,tptr))
			return -1;
	}

	if(format == 1) {
		if(ptr+2 > buf+tptr->offset+tptr->size)
			return -1;
		count = (*ptr++)<<8;
		count += (*ptr++);

		for(i=0; i<count; i++) {
			if(ptr+4 > buf+tptr->offset+tptr->size)
				return -1;

			offset = (*ptr++)<<8;
			offset += (*ptr++);

			if(offset) {
				if(verifyanchortable(tableoffset+offset,buf,tptr))
					return -1;
			}

			
			offset = (*ptr++)<<8;
			offset += (*ptr++);

			if(offset) {
				if(verifyanchortable(tableoffset+offset,buf,tptr))
					return -1;
			}
		}
	}
	else
		return -1;

	return 0;
}

int verifygposlookup4(unsigned int tableoffset,unsigned char *buf, table_type *tptr) 
{
	unsigned char *base=buf+tptr->offset+tableoffset;
	unsigned char *ptr = base;
	unsigned int format;
	unsigned int count;
	unsigned int i;
	unsigned int offset;
	unsigned int coverageoffset;
	unsigned char *bptr;

	if(ptr+2 > buf+tptr->offset+tptr->size)
		return -1;
	format = (*ptr++)<<8;
	format += (*ptr++);

	if(format == 1) {

		if(ptr+12 > buf+tptr->offset+tptr->size)
			return -1;
		coverageoffset = (*ptr++)<<8;
		coverageoffset += (*ptr++);

		if(coverageoffset) {
			if(verifycoveragetable(tableoffset+coverageoffset,buf,tptr))
				return -1;
		}

		coverageoffset = (*ptr++)<<8;
		coverageoffset += (*ptr++);

		if(coverageoffset) {
			if(verifycoveragetable(tableoffset+coverageoffset,buf,tptr))
				return -1;
		}

		count = (*ptr++)<<8;
		count += (*ptr++);

		offset = (*ptr++)<<8;
		offset += (*ptr++);

		if(offset) {
			if(verifymarkarray(tableoffset+offset,buf,tptr))
				return -1;
		}

		offset = (*ptr++)<<8;
		offset += (*ptr++);

		if(offset) {

			bptr = base + offset;

			count = (*bptr++)<<8;
			count += (*bptr++);

			for(i=0; i<count; i++) {
				if(ptr+2 > buf+tptr->offset+tptr->size)
					return -1;
				offset = (*bptr++)<<8;
				offset += (*bptr++);

				if(offset) {
					if(verifyanchortable(tableoffset,buf,tptr))
						return 0;
				}
			}
		} 
	}
	else
		return -1;
	
	return 0;
}


int verifygposlookup5(unsigned int tableoffset,unsigned char *buf, table_type *tptr) 
{
	unsigned char *base=buf+tptr->offset+tableoffset;
	unsigned char *ptr = base;
	unsigned int format;
	unsigned int count;
	unsigned int i;
	unsigned int j;
	unsigned int offset;
	unsigned int coverageoffset;
	unsigned char *lptr;
	unsigned char *laptr;
	unsigned int laoffset;
	unsigned int ancoffset;
	unsigned int ccount;

	if(ptr+2 > buf+tptr->offset+tptr->size)
		return -1;

	format = (*ptr++)<<8;
	format += (*ptr++);

	if(format == 1) {
		if(ptr+10 > buf+tptr->offset+tptr->size)
			return -1;

		coverageoffset = (*ptr++)<<8;
		coverageoffset += (*ptr++);

		if(coverageoffset) {
			if(verifycoveragetable(tableoffset+coverageoffset,buf,tptr))
				return -1;
		}

		coverageoffset = (*ptr++)<<8;
		coverageoffset += (*ptr++);

		if(coverageoffset) {
			if(verifycoveragetable(tableoffset+coverageoffset,buf,tptr))
				return -1;
		}

	
		count = (*ptr++)<<8;
		count += (*ptr++);

		offset = (*ptr++)<<8;
		offset += (*ptr++);

		if(offset) {
			if(verifymarkarray(tableoffset+offset,buf,tptr))
				return -1;
		}

		offset = (*ptr++)<<8;
		offset += (*ptr++);

		if(offset) {

			lptr = base + offset;

			if(lptr+2 > buf+tptr->offset+tptr->size)
				return -1;

			count = (*lptr++)<<8;
			count += (*lptr++);

			for(i=0; i<count; i++) {
				if(lptr+2 > buf+tptr->offset+tptr->size)
					return -1;
				laoffset = (*lptr++)<<8;
				laoffset += (*lptr++);

				if(!laoffset)
					continue;

				laptr = base + offset + laoffset;

				if(laptr+2 > buf+tptr->offset+tptr->size)
					return -1;

				ccount = (*laptr++)<<8;
				ccount += (*laptr++);

				for(j=0; j<ccount; j++) {
					if(laptr+2 > buf+tptr->offset+tptr->size)
						return -1;
					ancoffset = (*laptr++)<<8;
					ancoffset += (*laptr++);

					if(!ancoffset)
						continue;

					if(verifyanchortable(tableoffset + offset + laoffset + ancoffset,buf,tptr))
						return -1;
				} 
			} 
		}
	}
	else
		return -1;

	return 0;
}

int verifygposlookup6(unsigned int tableoffset,unsigned char *buf, table_type *tptr) 
{
	unsigned char *base=buf+tptr->offset+tableoffset;
	unsigned char *ptr = base;
	unsigned int format;
	unsigned int count;
	unsigned int i;
	unsigned int offset;
	unsigned int coverageoffset;
	unsigned char *m2ptr;
	unsigned int m2offset;

	if(ptr+2 > buf+tptr->offset+tptr->size)
		return -1;

	format = (*ptr++)<<8;
	format += (*ptr++);

	if(format == 1) {

		if(ptr+10 > buf+tptr->offset+tptr->size)
			return -1;

		coverageoffset = (*ptr++)<<8;
		coverageoffset += (*ptr++);

		if(coverageoffset) {
			if(verifycoveragetable(tableoffset+coverageoffset,buf,tptr))
				return -1;
		}

		coverageoffset = (*ptr++)<<8;
		coverageoffset += (*ptr++);

		if(coverageoffset) {
			if(verifycoveragetable(tableoffset+coverageoffset,buf,tptr))
				return -1;
		}

	
		count = (*ptr++)<<8;
		count += (*ptr++);

		offset = (*ptr++)<<8;
		offset += (*ptr++);

		if(offset) {
			if(verifymarkarray(tableoffset+offset,buf,tptr))
				return -1;
		}

		offset = (*ptr++)<<8;
		offset += (*ptr++);

		if(offset) {

			m2ptr = base + offset;

			if(m2ptr+2 > buf+tptr->offset+tptr->size)
				return -1;
		
			count = (*m2ptr++)<<8;
			count += (*m2ptr++);

			for(i=0; i<count; i++) {
				if(m2ptr+2 > buf+tptr->offset+tptr->size)
					return -1;
				m2offset = (*m2ptr++)<<8;
				m2offset += (*m2ptr++);

				if(m2offset) {
					if(verifyanchortable(tableoffset+offset+m2offset,buf,tptr))
						return -1;
				}
			}
		}
	}
	else
		return -1;

	return 0;
}


int verifygposlookup7(unsigned int tableoffset,unsigned char *buf, table_type *tptr) 
{
	unsigned char *base=buf+tptr->offset+tableoffset;
	unsigned char *ptr = base;
	unsigned int format;
	unsigned int count;
	unsigned int i;
	unsigned int j;
	unsigned int offset;
	unsigned int coverageoffset;
	unsigned int classdefoffset;
	unsigned char *prsptr;
	unsigned char *prptr;
	unsigned int proffset;
	unsigned int prcount;
	unsigned int glyphcount;
	unsigned int poscount;

	if(ptr+2 > buf+tptr->offset+tptr->size)
		return -1;

	format = (*ptr++)<<8;
	format += (*ptr++);

	if(format == 1) {
		if(ptr+4 > buf+tptr->offset+tptr->size)
			return -1;

		coverageoffset = (*ptr++)<<8;
		coverageoffset += (*ptr++);

		if(coverageoffset) {
			if(verifycoveragetable(tableoffset+coverageoffset,buf,tptr))
				return -1;
		}

		
		count = (*ptr++)<<8;
		count += (*ptr++);

		for(i=0; i<count; i++) {
			if(ptr+2 > buf+tptr->offset+tptr->size)
				return -1;

			offset = (*ptr++)<<8;
			offset += (*ptr++);

			if(!offset)
				continue;

			prsptr = base + offset;

			if(prsptr+2 > buf+tptr->offset+tptr->size)
				return -1;

			prcount = (*prsptr++)<<8;
			prcount += (*prsptr++);

			for(j=0; j<prcount; j++) {
				if(prsptr+2 > buf+tptr->offset+tptr->size)
					return -1;
				proffset = (*prsptr++)<<8;
				proffset += (*prsptr++);

				if(!proffset)
					continue;

				prptr = base + offset + proffset;

				if(prptr+4 > buf+tptr->offset+tptr->size)
					return -1;

				glyphcount = (*prptr++)<<8;
				glyphcount += (*prptr++);

				poscount = (*prptr++)<<8;
				poscount += (*prptr++);

				if(prptr + (glyphcount-1)*2 + poscount * 4 > buf+tptr->offset + tptr->size)
					return -1;
			}
		}
	}
	else if(format == 2) {

		if(ptr+6 > buf+tptr->offset+tptr->size)
			return -1;

		coverageoffset = (*ptr++)<<8;
		coverageoffset += (*ptr++);

		if(coverageoffset) {
			if(verifycoveragetable(tableoffset+coverageoffset,buf,tptr))
				return -1;
		}

		classdefoffset = (*ptr++)<<8;
		classdefoffset += (*ptr++);

		if(classdefoffset) {
			if(verifyclassdeftable(tableoffset+classdefoffset,buf,tptr))
				return -1;
		}

		
		count = (*ptr++)<<8;
		count += (*ptr++);

		for(i=0; i<count; i++) {
			if(ptr+2 > buf+tptr->offset+tptr->size)
				return -1;

			offset = (*ptr++)<<8;
			offset += (*ptr++);

			if(!offset)
				continue;

			prsptr = base + offset;
			if(prsptr+2 > buf+tptr->offset+tptr->size)
				return -1;

			prcount = (*prsptr++)<<8;
			prcount += (*prsptr++);

			for(j=0; j<prcount; j++) {
				if(prsptr+2 > buf+tptr->offset+tptr->size)
					return -1;
				proffset = (*prsptr++)<<8;
				proffset += (*prsptr++);

				if(!offset)
					continue;

				prptr = base + offset + proffset;

				if(prptr+4 > buf+tptr->offset+tptr->size)
					return -1;

				glyphcount = (*prptr++)<<8;
				glyphcount += (*prptr++);

				poscount = (*prptr++)<<8;
				poscount += (*prptr++);

				if(prptr + (glyphcount-1)*2 + poscount * 4 > buf+tptr->offset + tptr->size)
					return -1;
			}
		}
	}

	else if(format == 3) {
		if(ptr+4 > buf+tptr->offset+tptr->size)
			return -1;
		glyphcount = (*ptr++)<<8;
		glyphcount += (*ptr++);

		poscount = (*ptr++)<<8;
		poscount += (*ptr++);

		for(i=0; i<glyphcount; i++) {
			if(ptr+2 > buf+tptr->offset+tptr->size)
				return -1;
			coverageoffset = (*ptr++)<<8;
			coverageoffset += (*ptr++);

			if(coverageoffset) {
				if(verifycoveragetable(tableoffset+coverageoffset,buf,tptr))
					return -1;
			}
		}

		if(ptr + poscount * 4 > buf + tptr->offset + tptr->size)
			return -1;
	}

	else
		return -1;

	return 0;
}


int verifygposlookup8(unsigned int tableoffset,unsigned char *buf, table_type *tptr) 
{
	unsigned char *base=buf+tptr->offset+tableoffset;
	unsigned char *ptr = base;
	unsigned int format;
	unsigned int count;
	unsigned int i;
	unsigned int j;
	unsigned int offset;
	unsigned int coverageoffset;
	unsigned int classdefoffset;
	unsigned char *cprsptr;
	unsigned char *cprptr;
	unsigned int cproffset;
	unsigned int cprcount;
	unsigned int poscount;
	unsigned int lacount;
	unsigned int igcount;
	unsigned int bkcount;

	if(ptr+2 > buf+tptr->offset+tptr->size)
		return -1;

	format = (*ptr++)<<8;
	format += (*ptr++);

	if(format == 1) {

		if(ptr+4 > buf+tptr->offset+tptr->size)
			return -1;

		coverageoffset = (*ptr++)<<8;
		coverageoffset += (*ptr++);

		if(coverageoffset) {
			if(verifycoveragetable(tableoffset+coverageoffset,buf,tptr))
				return -1;
		}

		
		count = (*ptr++)<<8;
		count += (*ptr++);

		for(i=0; i<count; i++) {
			if(ptr+4 > buf+tptr->offset+tptr->size)
				return -1;

			offset = (*ptr++)<<8;
			offset += (*ptr++);

			if(!offset)
				continue;

			cprsptr = base + offset;

			if(cprsptr+2 > buf+tptr->offset+tptr->size)
				return -1;

			cprcount = (*cprsptr++)<<8;
			cprcount += (*cprsptr++);

			for(j=0; j<cprcount; j++) {
				if(cprsptr+2 > buf+tptr->offset+tptr->size)
					return -1;
				
				cproffset = (*cprsptr++)<<8;
				cproffset += (*cprsptr++);

				if(!cproffset)
					continue;

				cprptr = base + offset + cproffset;
				
				if(cprptr+2 > buf+tptr->offset+tptr->size)
					return -1;

				bkcount = (*cprptr++)<<8;
				bkcount += (*cprptr++);

				if(cprptr + bkcount *2 + 2 > buf + tptr->offset+tptr->size)
					return -1;

				igcount = (*cprptr++)<<8;
				igcount += (*cprptr++);

				if(cprptr + igcount *2+2 > buf + tptr->offset+tptr->size)
					return -1;

				lacount = (*cprptr++)<<8;
				lacount += (*cprptr++);

				if(cprptr + lacount *2+2 > buf + tptr->offset+tptr->size)
					return -1;

				poscount = (*cprptr++)<<8;
				poscount += (*cprptr++);

				if(cprptr + poscount *4 > buf + tptr->offset+tptr->size)
					return -1;
			}
		}
	}
	else if(format == 2) {

		if(ptr+10 > buf+tptr->offset+tptr->size)
			return -1;

		coverageoffset = (*ptr++)<<8;
		coverageoffset += (*ptr++);

		if(coverageoffset) {
			if(verifycoveragetable(tableoffset+coverageoffset,buf,tptr))
				return -1;
		}

		/* Backtrack Class Def */
		classdefoffset = (*ptr++)<<8;
		classdefoffset += (*ptr++);

		if(classdefoffset) {

			if(verifyclassdeftable(tableoffset+classdefoffset,buf,tptr))
				return -1;
		}

		/* Input Class Def */
		classdefoffset = (*ptr++)<<8;
		classdefoffset += (*ptr++);

		if(classdefoffset) {
			if(verifyclassdeftable(tableoffset+classdefoffset,buf,tptr))
				return -1;
		}

		/* Look ahead Class Def */
		classdefoffset = (*ptr++)<<8;
		classdefoffset += (*ptr++);

		if(classdefoffset) {
			if(verifyclassdeftable(tableoffset+classdefoffset,buf,tptr))
				return -1;
		}

		
		count = (*ptr++)<<8;
		count += (*ptr++);

		for(i=0; i<count; i++) {
			if(ptr+2 > buf+tptr->offset+tptr->size)
				return -1;

			offset = (*ptr++)<<8;
			offset += (*ptr++);

			if(!offset)
				continue;

			cprsptr = base + offset;

			if(cprsptr+2 > buf+tptr->offset+tptr->size)
				return -1;

			cprcount = (*cprsptr++)<<8;
			cprcount += (*cprsptr++);

			for(j=0; j<cprcount; j++) {
				if(cprsptr+2 > buf+tptr->offset+tptr->size)
					return -1;
				
				cproffset = (*cprsptr++)<<8;
				cproffset += (*cprsptr++);

				if(!cproffset)
					continue;

				cprptr = base + offset + cproffset;
				if(cprptr+8 > buf+tptr->offset+tptr->size)
					return -1;

				bkcount = (*cprptr++)<<8;
				bkcount += (*cprptr++);

				cprptr += bkcount*2;

				if(cprptr + 2 > buf + tptr->offset+tptr->size)
					return -1;

				igcount = (*cprptr++)<<8;
				igcount += (*cprptr++);

				cprptr += 2*(igcount-1);

				if(cprptr + 2 > buf + tptr->offset+tptr->size)
					return -1;

				lacount = (*cprptr++)<<8;
				lacount += (*cprptr++);

				cprptr += lacount*2;

				if(cprptr + 2 > buf + tptr->offset+tptr->size)
					return -1;

				poscount = (*cprptr++)<<8;
				poscount += (*cprptr++);

				

				if(cprptr + poscount *4 > buf + tptr->offset+tptr->size)
					return -1;
			}
		}
	}
	else if(format == 3) {
		if(ptr+2 > buf+tptr->offset+tptr->size)
			return -1;

		bkcount = (*ptr++)<<8;
		bkcount += (*ptr++);

		for(i=0; i<bkcount; i++) {
			
			if(ptr+2 > buf+tptr->offset+tptr->size)
				return -1;

			coverageoffset = (*ptr++)<<8;
			coverageoffset += (*ptr++);

			if(!coverageoffset)
				continue;

			if(verifycoveragetable(tableoffset+coverageoffset,buf,tptr))
				return -1;
		}

		if(ptr+2 > buf+tptr->offset+tptr->size)
			return -1;

		igcount = (*ptr++)<<8;
		igcount += (*ptr++);

		for(i=0; i<igcount; i++) {
			if(ptr+2 > buf+tptr->offset+tptr->size)
				return -1;

			coverageoffset = (*ptr++)<<8;
			coverageoffset += (*ptr++);

			if(!coverageoffset)
				continue;

			if(verifycoveragetable(tableoffset+coverageoffset,buf,tptr))
				return -1;
		}

		if(ptr+2 > buf+tptr->offset+tptr->size)
			return -1;

		lacount = (*ptr++)<<8;
		lacount += (*ptr++);

		for(i=0; i<lacount; i++) {

			if(ptr+2 > buf+tptr->offset+tptr->size)
				return -1;
			coverageoffset = (*ptr++)<<8;
			coverageoffset += (*ptr++);

			if(!coverageoffset)
				continue;

			if(verifycoveragetable(tableoffset+coverageoffset,buf,tptr))
				return -1;
		}

		if(ptr+2 > buf+tptr->offset+tptr->size)
			return -1;
	
		poscount = (*ptr++)<<8;
		poscount += (*ptr++);

		if(ptr + poscount *4 > buf + tptr->offset+tptr->size)
			return -1;
	}

	else
		return -1;

	return 0;
}

	
int verifygposlookup9(unsigned int tableoffset,unsigned char *buf, table_type *tptr) 
{
	unsigned char *base=buf+tptr->offset+tableoffset;
	unsigned char *ptr = base;
	unsigned int format;
	unsigned int offset;
	unsigned int lookuptype;
	
	if(ptr+2 > buf+tptr->offset+tptr->size)
		return -1;

	format = (*ptr++)<<8;
	format += (*ptr++);

	
	if(format == 1) {
		if(ptr+6 > buf+tptr->offset+tptr->size)
			return -1;
		
		lookuptype = (*ptr++)<<8;
		lookuptype += (*ptr++);

		offset = (*ptr++)<<24;
		offset += (*ptr++)<<16;
		offset += (*ptr++)<<8;
		offset += (*ptr++);

		

		if(lookuptype == 1) {
			if(verifygposlookup1(tableoffset + offset,buf,tptr))
				return -1;
		}
		else if(lookuptype == 2) {
			if(verifygposlookup2(tableoffset + offset,buf,tptr))
				return -1;
		}
		else if(lookuptype == 3) {
			if(verifygposlookup3(tableoffset + offset,buf,tptr))
				return -1;
		}
		else if(lookuptype == 4) {
			if(verifygposlookup4(tableoffset + offset,buf,tptr))
				return -1;
		}
		else if(lookuptype == 5) {
			if(verifygposlookup5(tableoffset + offset,buf,tptr))
				return -1;
		}
		else if(lookuptype == 6) {
			if(verifygposlookup6(tableoffset + offset,buf,tptr))
				return -1;
		}
		else if(lookuptype == 7) {
			if(verifygposlookup7(tableoffset + offset,buf,tptr))
				return -1;
		}
		else if(lookuptype == 8) {
			if(verifygposlookup8(tableoffset + offset,buf,tptr))
				return -1;
		}
		else if(lookuptype == 9) {
			return -1;
		}


	}
	else
		return -1;

	return 0;
}

int verifygposlookuplist(unsigned int tableoffset,unsigned char *buf, table_type *tptr) 
{
	unsigned int i;
	unsigned char *base=buf+tptr->offset+tableoffset;
	unsigned char *ptr = base;
	unsigned int offset;
	unsigned int count;
	unsigned int lookuptype;
	unsigned int stcount;
	unsigned int stoffset;
	unsigned char *ltptr;
	unsigned int j;

	if(ptr+2 > buf+tptr->offset+tptr->size){
		return -1;
	}
		

	count = (*ptr++)<<8;
	count += (*ptr++);

	

	for(i=0; i<count; i++) {

		if(ptr+2 > buf+tptr->offset+tptr->size){
			return -1;
		}
		
		offset = (*ptr++)<<8;
		offset += (*ptr++);

		if(!offset)
			continue;

		ltptr = base + offset;

		if(ltptr+6 > buf+tptr->offset+tptr->size) {
			return -1;
		} 
	
		lookuptype = (*ltptr++)<<8;
		lookuptype += (*ltptr++);

		ltptr += 2;

		stcount = (*ltptr++)<<8;
		stcount += (*ltptr++);

		for(j=0; j<stcount; j++) {
			if(ltptr+2 > buf+tptr->offset+tptr->size) {
				return -1;
			} 
			
			stoffset = (*ltptr++)<<8;
			stoffset += (*ltptr++);

			if(!stoffset)
				continue;

			if(lookuptype == 1) {
				if(verifygposlookup1(tableoffset+stoffset + offset,buf,tptr))
					return -1;
			}
			else if(lookuptype == 2) {
				if(verifygposlookup2(tableoffset+stoffset + offset,buf,tptr))
					return -1;
			}
			else if(lookuptype == 3) {
				if(verifygposlookup3(tableoffset+stoffset + offset,buf,tptr))
					return -1;
			}
			else if(lookuptype == 4) {
				if(verifygposlookup4(tableoffset+stoffset + offset,buf,tptr))
					return -1;
			}
			else if(lookuptype == 5) {
				if(verifygposlookup5(tableoffset+stoffset + offset,buf,tptr))
					return -1;
			}
			else if(lookuptype == 6) {
				if(verifygposlookup6(tableoffset+stoffset + offset,buf,tptr))
					return -1;
			}
			else if(lookuptype == 7) {
				if(verifygposlookup7(tableoffset+stoffset + offset,buf,tptr))
					return -1;
			}
			else if(lookuptype == 8) {
				if(verifygposlookup8(tableoffset+stoffset + offset,buf,tptr))
					return -1;
			}
			else if(lookuptype == 9) {
				if(verifygposlookup9(tableoffset+stoffset + offset,buf,tptr))
					return -1;
			}
			else
				return -1;
		}
	}

	return 0;
}

int verifygsublookup1(unsigned int tableoffset,unsigned char *buf, table_type *tptr) 
{
	unsigned char *base=buf+tptr->offset+tableoffset;
	unsigned char *ptr = base;
	unsigned int format;
	unsigned int count;
	unsigned int coverageoffset;

	if(ptr+4 > buf+tptr->offset+tptr->size)
		return -1;

	format = (*ptr++)<<8;
	format += (*ptr++);

	coverageoffset = (*ptr++)<<8;
	coverageoffset += (*ptr++);

	if(coverageoffset) {
		if(verifycoveragetable(tableoffset+coverageoffset,buf,tptr))
			return -1;
	}

	if(format == 1) {
		if(ptr+2 > buf+tptr->offset+tptr->size)
			return -1;
	}
	else if(format == 2) {
		if(ptr+2 > buf+tptr->offset+tptr->size)
			return -1;
		count = (*ptr++)<<8;
		count += (*ptr++);

		if(ptr+count*2 > buf+tptr->offset+tptr->size)
			return -1;
	}
	else
		return -1;

	return 0;
}


int verifygsublookup2(unsigned int tableoffset,unsigned char *buf, table_type *tptr) 
{
	unsigned char *base=buf+tptr->offset+tableoffset;
	unsigned char *ptr = base;
	unsigned int format;
	unsigned int count;
	unsigned int coverageoffset;
	unsigned int i;
	unsigned char *sqptr;
	unsigned int glyphcount;
	unsigned int offset;

	if(ptr+4 > buf+tptr->offset+tptr->size)
		return -1;

	format = (*ptr++)<<8;
	format += (*ptr++);

	coverageoffset = (*ptr++)<<8;
	coverageoffset += (*ptr++);

	if(coverageoffset) {
		if(verifycoveragetable(tableoffset+coverageoffset,buf,tptr))
			return -1;
	}

	if(format == 1) {
		if(ptr+2 > buf+tptr->offset+tptr->size)
			return -1;

		count = (*ptr++)<<8;
		count += (*ptr++);

		for(i=0; i<count; i++) {
			if(ptr+2 > buf+tptr->offset+tptr->size)
				return -1;

			offset = (*ptr++)<<8;
			offset += (*ptr++);

			if(!offset)
				continue;

			sqptr = base + offset;

			if(sqptr+2 > buf+tptr->offset+tptr->size)
				return -1;

			glyphcount = (*sqptr++)<<8;
			glyphcount += (*sqptr++);

			if(sqptr + 2 * glyphcount > buf+tptr->offset+tptr->size)
				return -1;
		}

	}
	else
		return -1;

	return 0;
}

int verifygsublookup3(unsigned int tableoffset,unsigned char *buf, table_type *tptr) 
{
	unsigned char *base=buf+tptr->offset+tableoffset;
	unsigned char *ptr = base;
	unsigned int format;
	unsigned int count;
	unsigned int coverageoffset;
	unsigned int i;
	unsigned char *altptr;
	unsigned int glyphcount;
	unsigned int offset;

	if(ptr+4 > buf+tptr->offset+tptr->size)
		return -1;

	format = (*ptr++)<<8;
	format += (*ptr++);

	coverageoffset = (*ptr++)<<8;
	coverageoffset += (*ptr++);

	if(coverageoffset) {
		if(verifycoveragetable(tableoffset+coverageoffset,buf,tptr))
			return -1;
	}

	if(format == 1) {

		if(ptr+2 > buf+tptr->offset+tptr->size)
			return -1;
		
		count = (*ptr++)<<8;
		count += (*ptr++);

		for(i=0; i<count; i++) {
			if(ptr+2 > buf+tptr->offset+tptr->size)
				return -1;

			offset = (*ptr++)<<8;
			offset += (*ptr++);

			if(!offset)
				continue;

			altptr = base + offset;

			if(altptr+2 > buf+tptr->offset+tptr->size)
				return -1;

			glyphcount = (*altptr++)<<8;
			glyphcount += (*altptr++);

			if(altptr + 2 * glyphcount > buf+tptr->offset+tptr->size)
				return -1;
		}

	}
	else
		return -1;

	return 0;
}




int verifygsublookup4(unsigned int tableoffset,unsigned char *buf, table_type *tptr) 
{
	unsigned char *base=buf+tptr->offset+tableoffset;
	unsigned char *ptr = base;
	unsigned int format;
	unsigned int count;
	unsigned int coverageoffset;
	unsigned int i;
	unsigned int j;
	unsigned char *ligptr;
	unsigned char *ligsetptr;
	unsigned int compcount;
	unsigned int offset;
	unsigned int ligoffset;
	unsigned int ligcount;

	if(ptr+4 > buf+tptr->offset+tptr->size)
		return -1;

	format = (*ptr++)<<8;
	format += (*ptr++);

	coverageoffset = (*ptr++)<<8;
	coverageoffset += (*ptr++);

	if(coverageoffset) {
		if(verifycoveragetable(tableoffset+coverageoffset,buf,tptr))
			return -1;
	}

	if(format == 1) {
		
		if(ptr+2 > buf+tptr->offset+tptr->size)
			return -1;

		count = (*ptr++)<<8;
		count += (*ptr++);

		for(i=0; i<count; i++) {
			if(ptr+2 > buf+tptr->offset+tptr->size)
				return -1;

			offset = (*ptr++)<<8;
			offset += (*ptr++);

			if(!offset)
				continue;

			ligsetptr = base + offset;

			if(ligsetptr+2 > buf+tptr->offset+tptr->size)
				return -1;

			ligcount = (*ligsetptr++)<<8;
			ligcount += (*ligsetptr++);

			for(j=0; j<ligcount; j++) {
				if(ligsetptr+2 > buf+tptr->offset+tptr->size)
					return -1;

				ligoffset = (*ligsetptr++)<<8;
				ligoffset += (*ligsetptr++);

				if(!ligoffset)
					continue;

				ligptr = base + offset + ligoffset;

				ligptr += 2;

				if(ligptr+2 > buf+tptr->offset+tptr->size)
					return -1;

				compcount = (*ligptr++)<<8;
				compcount += (*ligptr++);

				if(ligptr + 2 * (compcount-1) > buf+tptr->offset+tptr->size)
					return -1;
			}
		}

	}
	else
		return -1;

	return 0;
}

int verifygsublookup5(unsigned int tableoffset,unsigned char *buf, table_type *tptr) 
{
	unsigned char *base=buf+tptr->offset+tableoffset;
	unsigned char *ptr = base;
	unsigned int format;
	unsigned int count;
	unsigned int coverageoffset;
	unsigned int i;
	unsigned int j;
	unsigned char *subrulesetptr;
	unsigned char *subruleptr;
	unsigned int substcount;
	unsigned int glyphcount;
	unsigned int offset;
	unsigned int subruleoffset;
	unsigned int subrulecount;

	if(ptr+2 > buf+tptr->offset+tptr->size)
		return -1;

	format = (*ptr++)<<8;
	format += (*ptr++);

	if(format == 1) {

		if(ptr+4 > buf+tptr->offset+tptr->size)
			return -1;

		coverageoffset = (*ptr++)<<8;
		coverageoffset += (*ptr++);

		if(coverageoffset) {
			if(verifycoveragetable(tableoffset+coverageoffset,buf,tptr))
				return -1;
		}
		
		count = (*ptr++)<<8;
		count += (*ptr++);

		for(i=0; i<count; i++) {

			if(ptr+2 > buf+tptr->offset+tptr->size)
				return -1;

			offset = (*ptr++)<<8;
			offset += (*ptr++);

			if(!offset)
				continue;

			subrulesetptr = base + offset;

			if(subrulesetptr+2 > buf+tptr->offset+tptr->size)
				return -1;

			subrulecount = (*subrulesetptr++)<<8;
			subrulecount += (*subrulesetptr++);

			for(j=0; j<subrulecount; j++) {
				if(subrulesetptr+2 > buf+tptr->offset+tptr->size)
					return -1;

				subruleoffset = (*subrulesetptr++)<<8;
				subruleoffset += (*subrulesetptr++);

				if(!subruleoffset)
					continue;

				subruleptr = base + offset + subruleoffset;

				if(subruleptr+4 > buf+tptr->offset+tptr->size)
					return -1;

				glyphcount = (*subruleptr++)<<8;
				glyphcount += (*subruleptr++);

				substcount = (*subruleptr++)<<8;
				substcount += (*subruleptr++);

				if(subruleptr+2*glyphcount+4*substcount > buf+tptr->offset+tptr->size)
					return -1;
			}
		}

	}
	else if(format == 2) {
		if(ptr+6 > buf+tptr->offset+tptr->size)
			return -1;

		coverageoffset = (*ptr++)<<8;
		coverageoffset += (*ptr++);

		if(coverageoffset) {
			if(verifycoveragetable(tableoffset+coverageoffset,buf,tptr))
				return -1;
		}
		
		offset = (*ptr++)<<8;
		offset += (*ptr++);

		if(offset) {
			if(verifyclassdeftable(offset+tableoffset,buf,tptr))
				return -1;
		}

		count = (*ptr++)<<8;
		count += (*ptr++);

		for(i=0; i<count; i++) {
			if(ptr+2 > buf+tptr->offset+tptr->size)
				return -1;

			offset = (*ptr++)<<8;
			offset += (*ptr++);

			if(!offset)
				continue;

			subrulesetptr = base + offset;
			if(subrulesetptr+2 > buf+tptr->offset+tptr->size)
				return -1;

			subrulecount = (*subrulesetptr++)<<8;
			subrulecount += (*subrulesetptr++);

			for(j=0; j<subrulecount; j++) {
				if(subrulesetptr+2 > buf+tptr->offset+tptr->size)
					return -1;
				
				subruleoffset = (*subrulesetptr++)<<8;
				subruleoffset += (*subrulesetptr++);

				if(!subruleoffset)
					continue;

				subruleptr = base + offset + subruleoffset;

				if(subruleptr+4 > buf+tptr->offset+tptr->size)
					return -1;

				glyphcount = (*subruleptr++)<<8;
				glyphcount += (*subruleptr++);

				substcount = (*subruleptr++)<<8;
				substcount += (*subruleptr++);

				if(subruleptr+2*(glyphcount-1)+4*substcount > buf+tptr->offset+tptr->size)
					return -1;
			}
		}

	}
	else if(format == 3) {
		if(ptr+4 > buf+tptr->offset+tptr->size)
			return -1;
		glyphcount = (*ptr++)<<8;
		glyphcount += (*ptr++);

		substcount = (*ptr++)<<8;
		substcount += (*ptr++);

		for(i=0; i<glyphcount; i++) {
		
			if(ptr+2 > buf+tptr->offset+tptr->size)
				return -1;
			coverageoffset = (*ptr++)<<8;
			coverageoffset += (*ptr++);

			if(coverageoffset) {
				if(verifycoveragetable(tableoffset+coverageoffset,buf,tptr))
					return -1;
			}
		}

		if(ptr+4*substcount > buf+tptr->offset+tptr->size)
			return -1;
	}

	else
		return -1;

	return 0;
}

int verifygsublookup6(unsigned int tableoffset,unsigned char *buf, table_type *tptr) 
{
	unsigned char *base=buf+tptr->offset+tableoffset;
	unsigned char *ptr = base;
	unsigned int format;
	unsigned int count;
	unsigned int coverageoffset;
	unsigned int i;
	unsigned int j;
	unsigned char *chainptr;
	unsigned char *chainsetptr;
	unsigned int offset;
	unsigned int chainoffset;
	unsigned int chaincount;
	unsigned int glyphcount;

	if(ptr+2 > buf+tptr->offset+tptr->size)
		return -1;

	format = (*ptr++)<<8;
	format += (*ptr++);

	
	if(format == 1) {
		if(ptr+4 > buf+tptr->offset+tptr->size)
			return -1;

		coverageoffset = (*ptr++)<<8;
		coverageoffset += (*ptr++);

		if(coverageoffset) {
			if(verifycoveragetable(tableoffset+coverageoffset,buf,tptr))
				return -1;
		}

		
		count = (*ptr++)<<8;
		count += (*ptr++);

		for(i=0; i<count; i++) {
			if(ptr+2 > buf+tptr->offset+tptr->size)
				return -1;

			offset = (*ptr++)<<8;
			offset += (*ptr++);

			if(!offset)
				continue;

			chainsetptr = base + offset;

			if(chainsetptr+2 > buf+tptr->offset+tptr->size)
				return -1;

			chaincount = (*chainsetptr++)<<8;
			chaincount += (*chainsetptr++);

			for(j=0; j<chaincount; j++) {
				if(chainsetptr+2 > buf+tptr->offset+tptr->size)
					return -1;
				chainoffset = (*chainsetptr++)<<8;
				chainoffset += (*chainsetptr++);

				if(!chainoffset)
					continue;

				chainptr = base + offset + chainoffset;

				if(chainptr+2 > buf+tptr->offset+tptr->size)
					return -1;

				glyphcount = (*chainptr++)<<8;
				glyphcount += (*chainptr++);

				chainptr += glyphcount*2;

				if(chainptr+2 > buf+tptr->offset+tptr->size)
					return -1;

				
				glyphcount = (*chainptr++)<<8;
				glyphcount += (*chainptr++);

				chainptr += (glyphcount-1)*2;

				if(chainptr +2 > buf+tptr->offset+tptr->size)
					return -1;

				glyphcount = (*chainptr++)<<8;
				glyphcount += (*chainptr++);

				chainptr += glyphcount*2;
				if(chainptr +2  > buf+tptr->offset+tptr->size)
					return -1;

				glyphcount = (*chainptr++)<<8;
				glyphcount += (*chainptr++);

				if(chainptr  > buf+tptr->offset+tptr->size)
					return -1;

			}
		}
	}
	else if(format == 2) {
		if(ptr+10 > buf+tptr->offset+tptr->size)
			return -1;

		coverageoffset = (*ptr++)<<8;
		coverageoffset += (*ptr++);

		if(coverageoffset) {
			if(verifycoveragetable(tableoffset+coverageoffset,buf,tptr))
				return -1;
		}

		/* Backtrack Class Def */
		offset = (*ptr++)<<8;
		offset += (*ptr++);

		if(offset) {

			if(verifyclassdeftable(offset+tableoffset,buf,tptr))
				return -1;
		}

		/* Input Class Def */
		offset = (*ptr++)<<8;
		offset += (*ptr++);

		if(offset) {
			if(verifyclassdeftable(offset+tableoffset,buf,tptr))
				return -1;
		}

		/* Look Ahead Class Def */
		offset = (*ptr++)<<8;
		offset += (*ptr++);
		

		if(offset) {
			if(verifyclassdeftable(offset+tableoffset,buf,tptr))
				return -1;
		}


		count = (*ptr++)<<8;
		count += (*ptr++);

		for(i=0; i<count; i++) {

			if(ptr+2 > buf+tptr->offset+tptr->size)
				return -1;

			offset = (*ptr++)<<8;
			offset += (*ptr++);

			if(!offset)
				continue;

			chainsetptr = base + offset;

			if(chainsetptr+2 > buf+tptr->offset+tptr->size)
				return -1;

			chaincount = (*chainsetptr++)<<8;
			chaincount += (*chainsetptr++);

			for(j=0; j<chaincount; j++) {
				if(chainsetptr+2 > buf+tptr->offset+tptr->size)
					return -1;
				
				chainoffset = (*chainsetptr++)<<8;
				chainoffset += (*chainsetptr++);

				if(!chainoffset)
					continue;

				chainptr = base + offset + chainoffset;
				if(chainptr+2 > buf+tptr->offset+tptr->size)
					return -1;

				glyphcount = (*chainptr++)<<8;
				glyphcount += (*chainptr++);

				chainptr += glyphcount*2;
				if(chainptr+2 > buf+tptr->offset+tptr->size)
					return -1;


				glyphcount = (*chainptr++)<<8;
				glyphcount += (*chainptr++);

				chainptr += (glyphcount-1)*2;
				if(chainptr+2 > buf+tptr->offset+tptr->size)
					return -1;

				glyphcount = (*chainptr++)<<8;
				glyphcount += (*chainptr++);

				chainptr += glyphcount*2;
				if(chainptr+2 > buf+tptr->offset+tptr->size)
					return -1;

				glyphcount = (*chainptr++)<<8;
				glyphcount += (*chainptr++);

				chainptr += glyphcount*4;

				if(chainptr > buf+tptr->offset+tptr->size)
					return -1;

			}
		}
	}

	else if(format == 3) {
		if(ptr+8 > buf+tptr->offset+tptr->size)
			return -1;
		count = (*ptr++)<<8;
		count += (*ptr++);

		for(i=0; i<count; i++) {
			if(ptr+2 > buf+tptr->offset+tptr->size)
				return -1;
			offset = (*ptr++)<<8;
			offset += (*ptr++);

			if(offset) {
				if(verifycoveragetable(tableoffset+offset,buf,tptr))
					return -1;
			}
		}

		count = (*ptr++)<<8;
		count += (*ptr++);

		for(i=0; i<count; i++) {

			offset = (*ptr++)<<8;
			offset += (*ptr++);

			if(offset) {
				if(verifycoveragetable(tableoffset+offset,buf,tptr))
					return -1;
			}
		}

		count = (*ptr++)<<8;
		count += (*ptr++);

		for(i=0; i<count; i++) {
			offset = (*ptr++)<<8;
			offset += (*ptr++);

			if(offset) {
				if(verifycoveragetable(tableoffset+offset,buf,tptr))
					return -1;
			}
		}

		count = (*ptr++)<<8;
		count += (*ptr++);

		if(ptr+count*4 > buf+tptr->offset+tptr->size)
			return -1;
	}
	else
		return -1;
	return 0;
}


int verifygsublookup7(unsigned int tableoffset,unsigned char *buf, table_type *tptr) 
{
	unsigned char *base=buf+tptr->offset+tableoffset;
	unsigned char *ptr = base;
	unsigned int format;
	unsigned int offset;
	unsigned int lookuptype;
	
	if(ptr+2 > buf+tptr->offset+tptr->size)
		return -1;

	format = (*ptr++)<<8;
	format += (*ptr++);

	
	if(format == 1) {
		if(ptr+6 > buf+tptr->offset+tptr->size)
			return -1;

		lookuptype = (*ptr++)<<8;
		lookuptype += (*ptr++);

		offset = (*ptr++)<<24;
		offset += (*ptr++)<<16;
		offset += (*ptr++)<<8;
		offset += (*ptr++);

		

		if(lookuptype == 1) {
			if(verifygsublookup1(tableoffset + offset,buf,tptr))
				return -1;
		}
		else if(lookuptype == 2) {
			if(verifygsublookup2(tableoffset + offset,buf,tptr))
				return -1;
		}
		else if(lookuptype == 3) {
			if(verifygsublookup3(tableoffset + offset,buf,tptr))
				return -1;
		}
		else if(lookuptype == 4) {
			if(verifygsublookup4(tableoffset + offset,buf,tptr))
				return -1;
		}
		else if(lookuptype == 5) {
			if(verifygsublookup5(tableoffset + offset,buf,tptr))
				return -1;
		}
		else if(lookuptype == 6) {
			if(verifygsublookup6(tableoffset + offset,buf,tptr))
				return -1;
		}
		else if(lookuptype == 7) {
			return -1;
		}
		else if(lookuptype == 8) {
			if(verifygsublookup8(tableoffset + offset,buf,tptr))
			return -1;
		}


	}
	else
		return -1;

	return 0;
}

		
int verifygsublookup8(unsigned int tableoffset,unsigned char *buf, table_type *tptr) 
{
	unsigned char *base=buf+tptr->offset+tableoffset;
	unsigned char *ptr = base;
	unsigned int format;
	unsigned int count;
	unsigned int coverageoffset;
	unsigned int i;
	unsigned int offset;

	if(ptr+2 > buf+tptr->offset+tptr->size)
		return -1;

	format = (*ptr++)<<8;
	format += (*ptr++);

	
	if(format == 1) {
		if(ptr+4 > buf+tptr->offset+tptr->size)
			return -1;

		coverageoffset = (*ptr++)<<8;
		coverageoffset += (*ptr++);

		if(coverageoffset) {
			if(verifycoveragetable(tableoffset+coverageoffset,buf,tptr))
				return -1;
		}

		
		count = (*ptr++)<<8;
		count += (*ptr++);

		for(i=0; i<count; i++) {
			if(ptr+2 > buf+tptr->offset+tptr->size)
				return -1;
			offset = (*ptr++)<<8;
			offset += (*ptr++);

			if(offset) {
				if(verifycoveragetable(tableoffset+offset,buf,tptr))
					return -1;
			}
		}
		if(ptr+2 > buf+tptr->offset+tptr->size)
			return -1;

		count = (*ptr++)<<8;
		count += (*ptr++);

		for(i=0; i<count; i++) {
			if(ptr+2 > buf+tptr->offset+tptr->size)
				return -1;

			offset = (*ptr++)<<8;
			offset += (*ptr++);

			if(offset) {
				if(verifycoveragetable(tableoffset+offset,buf,tptr))
					return -1;
			}
		}

		if(ptr+2 > buf+tptr->offset+tptr->size)
			return -1;

		count = (*ptr++)<<8;
		count += (*ptr++);

		if(ptr+count*2 > buf+tptr->offset+tptr->size)
			return -1;
	}
	else
		return -1;
	return 0;
}



int verifygsublookuplist(unsigned int tableoffset,unsigned char *buf, table_type *tptr) 
{
	unsigned int i;
	unsigned char *base=buf+tptr->offset+tableoffset;
	unsigned char *ptr = base;
	unsigned int offset;
	unsigned int count;
	unsigned int lookuptype;
	unsigned int stcount;
	unsigned int stoffset;
	unsigned char *ltptr;
	unsigned int j;

	if(ptr+2 > buf+tptr->offset+tptr->size)
		return -1;

	count = (*ptr++)<<8;
	count += (*ptr++);

	

	for(i=0; i<count; i++) {
		if(ptr+2 > buf+tptr->offset+tptr->size)
			return -1;

		offset = (*ptr++)<<8;
		offset += (*ptr++);

		if(!offset)
			continue;

		ltptr = base + offset;
		
		if(ltptr+6 > buf+tptr->offset+tptr->size)
			return -1;

		lookuptype = (*ltptr++)<<8;
		lookuptype += (*ltptr++);

		ltptr += 2;

		stcount = (*ltptr++)<<8;
		stcount += (*ltptr++);

		for(j=0; j<stcount; j++) {
			if(ltptr+2 > buf+tptr->offset+tptr->size)
				return -1;
			stoffset = (*ltptr++)<<8;
			stoffset += (*ltptr++);

			if(!stoffset)
				continue;

			if(lookuptype == 1) {
				if(verifygsublookup1(tableoffset+stoffset + offset,buf,tptr))
					return -1;
			}
			else if(lookuptype == 2) {
				if(verifygsublookup2(tableoffset+stoffset + offset,buf,tptr))
					return -1;
			}
			else if(lookuptype == 3) {
				if(verifygsublookup3(tableoffset+stoffset + offset,buf,tptr))
					return -1;
			}
			else if(lookuptype == 4) {
				if(verifygsublookup4(tableoffset+stoffset + offset,buf,tptr))
					return -1;
			}
			else if(lookuptype == 5) {
				if(verifygsublookup5(tableoffset+stoffset + offset,buf,tptr))
					return -1;
			}
			else if(lookuptype == 6) {
				if(verifygsublookup6(tableoffset+stoffset + offset,buf,tptr))
					return -1;
			}
			else if(lookuptype == 7) {
				if(verifygsublookup7(tableoffset+stoffset + offset,buf,tptr))
					return -1;
			}
			else if(lookuptype == 8) {
				if(verifygsublookup8(tableoffset+stoffset + offset,buf,tptr))
					return -1;
			}
		}
	}

	return 0;
}





int verifyfeaturelist(unsigned int tableoffset,unsigned char *buf, table_type *tptr) 
{
	unsigned int i;
	unsigned char *base=buf+tptr->offset+tableoffset;
	unsigned char *ptr = base;
	unsigned int offset;
	unsigned int count;
	unsigned int lookupcount;
	unsigned char *ftptr;

	if(ptr+2 > buf+tptr->offset+tptr->size)
		return -1;

	count = (*ptr++)<<8;
	count += (*ptr++);

	for(i=0; i<count; i++) {
		ptr += 4;

		if(ptr+2 > buf+tptr->offset+tptr->size)
			return -1;

		offset = (*ptr++)<<8;
		offset += (*ptr++);
		
		if(!offset)
			continue;

		ftptr = base + offset;

		ftptr += 2;

		if(ftptr+2 > buf+tptr->offset+tptr->size)
			return -1;

		lookupcount = (*ftptr++)<<8;
		lookupcount += (*ftptr++);

		if(ftptr + lookupcount*2 > buf + tptr->offset+tptr->size)
			return -1;
	}

	return 0;
}




int verifylangsystable(unsigned int tableoffset,unsigned char *buf, table_type *tptr) 
{
	unsigned char *base=buf+tptr->offset+tableoffset;
	unsigned char *ptr = base;
	unsigned int count;
	
	ptr += 4;

	if(ptr+2 > buf+tptr->offset+tptr->size)
		return -1;

	count = (*ptr++)<<8;
	count += (*ptr++);

	if(ptr + count*2 > buf + tptr->offset+tptr->size)
		return -1;

	return 0;
}


		
int verifyscriptlist(unsigned int tableoffset,unsigned char *buf, table_type *tptr) 
{
	unsigned int i;
	unsigned char *base=buf+tptr->offset+tableoffset;
	unsigned char *ptr = base;
	unsigned int offset;
	unsigned int lsoffset;
	unsigned int count;
	unsigned int langcount;
	unsigned char *stptr;
	unsigned int j;

	if(ptr+2 > buf+tptr->offset+tptr->size)
		return -1;

	count = (*ptr++)<<8;
	count += (*ptr++);

	for(i=0; i<count; i++) {
		ptr += 4;

		if(ptr+2 > buf+tptr->offset+tptr->size)
			return -1;
		offset = (*ptr++)<<8;
		offset += (*ptr++);

		if(!offset)
			continue;
		
		stptr = base + offset;

		if(stptr+4 > buf+tptr->offset+tptr->size)
			return -1;

		/* default language system */
		lsoffset = (*stptr++)<<8;
		lsoffset += (*stptr++);

		if(verifylangsystable(lsoffset+offset+tableoffset,buf,tptr))
			return -1;

		langcount = (*stptr++)<<8;
		langcount += (*stptr++);

		for(j=0; j<langcount; j++) {
			stptr += 4;

			if(stptr+2 > buf+tptr->offset+tptr->size)
				return -1;
			lsoffset = (*stptr++)<<8;
			lsoffset += (*stptr++);

			if(lsoffset) {

				if(verifylangsystable(lsoffset+offset+tableoffset,buf,tptr))
					return -1;
			}
		}

	}

	return 0;
}


int verifygsubtable(unsigned char *buf, truetype_type *tt)
{
	table_type *tptr = gettable(tt,"GSUB");
	unsigned int i;
	unsigned char *ptr;
	unsigned long checksum = 0;
	unsigned int offset;

	
	if(!tptr) {
		return 0;
	}

	ptr=buf+tptr->offset;

	if(tptr->offset + tptr->size > tt->filelen)	{
		tt->errors[tt->numberOfErrors] = ERRORINTABLEGSUB;
		tt->numberOfErrors++;
		return 0;
	}  

	for(i=0; i<(tptr->size+3)>>2; i++) {
		
		checksum += (*ptr++)<<24;
		checksum += (*ptr++)<<16;
		checksum += (*ptr++)<<8;
		checksum += (*ptr++);
		
	}

	if(checksum != tptr->checksum) {
		tt->errors[tt->numberOfErrors] = CHECKSUMERRORGSUB;
		tt->numberOfErrors++;
	}

	ptr = buf + tptr->offset + 4;

	if(ptr+6 > buf+tptr->offset+tptr->size){
		tt->errors[tt->numberOfErrors] = ERRORINTABLEGSUB;
		tt->numberOfErrors++;
		return -1;
	}
		

	offset = (*ptr++)<<8;
	offset += (*ptr++);

	if(offset) {
		if(verifyscriptlist(offset,buf,tptr)) {
			tt->errors[tt->numberOfErrors] = ERRORINTABLEGSUB;
			tt->numberOfErrors++;
			return 0;
		}
	}

	offset = (*ptr++)<<8;
	offset += (*ptr++);

	if(offset) {
		if(verifyfeaturelist(offset,buf,tptr)) {
			tt->errors[tt->numberOfErrors] = ERRORINTABLEGSUB;
			tt->numberOfErrors++;
			return 0;
	}	
	}

	offset = (*ptr++)<<8;
	offset += (*ptr++);

	if(offset) {
		if (verifygsublookuplist(offset,buf,tptr)) {
			tt->errors[tt->numberOfErrors] = ERRORINTABLEGSUB;
			tt->numberOfErrors++;
			return 0;
		}
	}

	return 0;
}

int verifygpostable(unsigned char *buf, truetype_type *tt)
{
	table_type *tptr = gettable(tt,"GPOS");
	unsigned int i;
	unsigned char *ptr;
	unsigned long checksum = 0;
	unsigned int offset;

	
	if(!tptr) {
		return 0;
	}
	ptr=buf+tptr->offset;

	if(tptr->offset + tptr->size > tt->filelen)	{
		tt->errors[tt->numberOfErrors] = ERRORINTABLEGPOS;
		tt->numberOfErrors++;
		return 0;
	}  

	for(i=0; i<(tptr->size+3)>>2; i++) {
		
		checksum += (*ptr++)<<24;
		checksum += (*ptr++)<<16;
		checksum += (*ptr++)<<8;
		checksum += (*ptr++);
		
	}

	if(checksum != tptr->checksum) {
		tt->errors[tt->numberOfErrors] = CHECKSUMERRORGPOS;
		tt->numberOfErrors++;
	}

	ptr = buf + tptr->offset + 4;

	if(ptr+6 > buf+tptr->offset+tptr->size){
		tt->errors[tt->numberOfErrors] = ERRORINTABLEGPOS;
		tt->numberOfErrors++;
		return -1;
	}

	offset = (*ptr++)<<8;
	offset += (*ptr++);

	if(offset) {

		if(verifyscriptlist(offset,buf,tptr)) {
			tt->errors[tt->numberOfErrors] = ERRORINTABLEGPOS;
			tt->numberOfErrors++;
			return 0;
		} 
	}

	offset = (*ptr++)<<8;
	offset += (*ptr++);

	if(offset) {


		if(verifyfeaturelist(offset,buf,tptr)) {
			tt->errors[tt->numberOfErrors] = ERRORINTABLEGPOS;
			tt->numberOfErrors++;
			return 0;
		} 
	}

	offset = (*ptr++)<<8;
	offset += (*ptr++);

	if(offset) {
		if (verifygposlookuplist(offset,buf,tptr)) {
			tt->errors[tt->numberOfErrors] = ERRORINTABLEGPOS;
			tt->numberOfErrors++;
			return 0;
		} 
	}

	return 0;
}
int verifyttfotf(unsigned char *buf, unsigned char *ptr, truetype_type *tt, int len)
{
	int i;
	table_type *t;
	int status;
	int otto = 0x4f54544f;
	int needcomma = 0;
	int badfont = 0;

	tt->numberOfErrors = 0;
	tt->numberOfErrors = 0;
	tt->level1 = 1;
	tt->level2 = 1;
	tt->level3 = 1;
	tt->level4 = 1;

	if(ptr+8 > buf+len){
		printf("File header error.\n");
		return -1;
	}
		
	tt->numtables = (*ptr++)<<8;
	tt->numtables += (*ptr++);

	tt->searchRange = (*ptr++)<<8;
	tt->searchRange += (*ptr++);

	tt->entrySelector = (*ptr++)<<8;
	tt->entrySelector += (*ptr++);

	tt->rangeShift = (*ptr++)<<8;
	tt->rangeShift += (*ptr++);

	tt->tables = (table_type *)malloc(tt->numtables*sizeof(table_type));
	t = tt->tables;

	for(i=0; i<tt->numtables; i++) {
		if(ptr+16 > buf+len){
			tt->errors[tt->numberOfErrors] = FILECHECKSUMERROR;
			tt->numberOfErrors++;
			return -1;
		} 
		t->tag = (*ptr++);
		t->tag += (*ptr++)<<8;
		t->tag += (*ptr++)<<16;
		t->tag += (*ptr++)<<24;

		t->checksum = (*ptr++)<<24;
		t->checksum += (*ptr++)<<16;
		t->checksum += (*ptr++)<<8;
		t->checksum += (*ptr++);
	
		t->offset = (*ptr++)<<24;
		t->offset += (*ptr++)<<16;
		t->offset += (*ptr++)<<8;
		t->offset += (*ptr++);

		t->size = (*ptr++)<<24;
		t->size += (*ptr++)<<16;
		t->size += (*ptr++)<<8;
		t->size += (*ptr++);

		if(t->offset+t->size > (unsigned int)len) {
			printf("File header error.\n");
			return 0;
		} 

		
		t++;
    
	} 

	status = verifyheadtable(buf,tt);
	if(status) {
		tt->level1 = tt->level3 = 0;
		goto error;
	}

    status = verifyhheatable(buf,tt);
	if(status) {
		tt->level1 = tt->level3 = 0;
		goto error;
	}

	status = verifymaxptable(buf,tt);
	if(status){
		tt->level1 = tt->level3 = 0;
		goto error;
	}

	verifyos2table(buf,tt);
	
	verifynametable(buf,tt);
	
	verifyposttable(buf,tt); 
	

	status = verifycmaptable(buf,tt);
	if(status) {
		tt->level1 = tt->level3 = 0;
		goto error;
	}

	verifyhmtxtable(buf,tt);
	
		
	if(tt->version == 0x00010000) {
		tt->level3 = 0;
		status = verifylocatable(buf,tt);
		if(status)
			goto error;
  
		status = verifyglyftable(buf,tt);
		if(status)
			goto error;
			
		verifycvttable(buf,tt);
	
		verifyfpgmtable(buf,tt);
	
		verifypreptable(buf,tt);
	}
	
	if(tt->version == otto) {
		tt->level1=0;
		verifycfftable(buf,tt);
		verifyvorgtable(buf,tt);
	}
	
	verifyeblcebdttable(buf,tt);
	
	verifyebsctable(buf,tt);
	
	  
	verifybasetable(buf,tt);
	
	verifygdeftable(buf,tt);
	
	verifygpostable(buf,tt);
	
	verifygsubtable(buf,tt);
	
	verifyjstftable(buf,tt);

error:

	for(i=0; i<tt->numberOfErrors; i++) {
		printf("%s\n",errorstring[tt->errors[i]]);
		if(tt->errors[i] > CHECKSUMERRORGLYF)
			badfont = 1;
	}

	if(badfont)
		return 0;

	printf("\nTables present:\n");

	if(tt->level1) {
		printf("Level 1\n");
		printf("cmap,head,hhea,hmtx,maxp,name,OS/2,post,loca,glyf\n\n");
	}

	if(tt->level3) {
		
		printf("Level 3\n");
		printf("cmap,head,hhea,hmtx,maxp,name,OS/2,post,cff");
		
		if(gettable(tt,"VORG"))
			printf(",VORG\n\n");
		else
			printf("\n\n");
	}

	if(gettable(tt,"prep") || gettable(tt,"cvt ") || gettable(tt,"fpgm")) {
		int count=-1;
		printf("Level 2\n");

		if(gettable(tt,"prep"))
			count++;
		if(gettable(tt,"cvt "))
			count++;
		if(gettable(tt,"fpgm"))
			count++;

		

		if(gettable(tt,"prep")) {
			printf("prep");

			if(count > 0) {
				printf(",");
				count--;
			} 
		} 

		if(gettable(tt,"cvt ")) {
			printf("cvt");
				if(count > 0) {
				printf(",");
				
			} 
		} 

		if(gettable(tt,"fpgm")) {
			printf("fpgm");
		}
		printf("\n\n");
	}
		
	if(gettable(tt,"BASE") || gettable(tt,"GDEF") || gettable(tt,"GPOS")
		 || gettable(tt,"GSUB") || gettable(tt,"JSTF") || gettable(tt,"EBLC")
		 || gettable(tt,"EBDT") || gettable(tt,"EBSC")) {
		int count=-1;
		printf("Level 4\n");

		if(gettable(tt,"BASE"))
			count++;
		if(gettable(tt,"GDEF"))
			count++;
		if(gettable(tt,"GPOS"))
			count++;
		if(gettable(tt,"GSUB"))
			count++;
		if(gettable(tt,"JSTF"))
			count++;
		if(gettable(tt,"EBLC"))
			count++;
		if(gettable(tt,"EBDT"))
			count++;
		if(gettable(tt,"EBSC"))
			count++;

		

		if(gettable(tt,"BASE")) {
			printf("BASE");

			if(count > 0) {
				printf(",");
				count--;
			} 
		} 

		if(gettable(tt,"GDEF")) {
			printf("GDEF");
				if(count > 0) {
				printf(",");
				count--;
				
			} 
		} 
		if(gettable(tt,"GPOS")) {
			printf("GPOS");

			if(count > 0) {
				printf(",");
				count--;
			} 
		} 

		if(gettable(tt,"GSUB")) {
			printf("GSUB");
			if(count > 0) {
				printf(",");
				count--;
				
			}  
		}  

		if(gettable(tt,"JSTF")) {
			printf("JSFT");
			if(count > 0) {
				printf(",");
				count--;
				
			}  
		}  

		if(gettable(tt,"EBLC")) {
			printf("EBLC");
			if(count > 0) {
				printf(",");
				count--;
				
			}  
		}  

		if(gettable(tt,"EBSC")) {
			printf("EBSC");
			if(count > 0) {
				printf(",");
				count--;
				
			}  
		}  

		if(gettable(tt,"EBDT")) {
			printf("EBDT");
		}
		printf("\n\n");
	}
	
	printf("Additional (optional) tables present:\n");
	t = tt->tables;

	for(i=0; i<tt->numtables; i++) {
		
		char text[5];
		unsigned char *p = (unsigned char *)&t->tag;
		if(istable(t,"head")) {
			t++;
			continue;
		}
		if(istable(t,"cmap")) {
			t++;
			continue;
		}
		if(istable(t,"hmtx")) {
			t++;
			continue;
		}
		if(istable(t,"post")) {
			t++;
			continue;
		}
		if(istable(t,"name")) {
			t++;
			continue;
		}
		if(istable(t,"OS/2")) {
			t++;
			continue;
		}
		if(istable(t,"maxp")) {
			t++;
			continue;
		}
		if(istable(t,"hhea")) {
			t++;
			continue;
		}
		if(istable(t,"loca")) {
			t++;
			continue;
		}
		if(istable(t,"glyf")) {
			t++;
			continue;
		}
		if(istable(t,"cvt ")) {
			t++;
			continue;
		}
		if(istable(t,"fpgm")) {
			t++;
			continue;
		}
		if(istable(t,"prep")) {
			t++;
			continue;
		}
		if(istable(t,"CFF ")) {
			t++;
			continue;
		}
		if(istable(t,"VORG")) {
			t++;
			continue;
		}
		if(istable(t,"EBSC")) {
			t++;
			continue;
		}
		if(istable(t,"EBLC")) {
			t++;
			continue;
		}
		if(istable(t,"EBDT")) {
			t++;
			continue;
		}
		if(istable(t,"JSTF")) {
			t++;
			continue;
		}
		if(istable(t,"GSUB")) {
			t++;
			continue;
		}
		if(istable(t,"GPOS")) {
			t++;
			continue;
		}
		if(istable(t,"GDEF")) {
			t++;
			continue;
		}
		if(istable(t,"BASE")) {
			t++;
			continue;
		}
		if(needcomma)
			printf(",");
		needcomma = 1;
		text[0] = *p++;
		text[1] = *p++;
		text[2] = *p++;
		text[3] = *p++;
		text[4] = 0;
		printf("%s",text);
		t++;
	}
	printf("\n");
	



	
	

	return 0;
}

int verifyttc(unsigned char *buf, unsigned char *ptr, truetype_type *tt, int len)
{
	unsigned int i;
	unsigned int version;
	unsigned int numfonts;
	unsigned int offset;
	unsigned char *vptr;
	


	if(ptr+8 > buf+len){
		printf("Ttc file header error.\n");
		return -1;
	}

	version = (*ptr++)<<24;
	version += (*ptr++)<<16;
	version += (*ptr++)<<8;
	version += (*ptr++);

	numfonts = (*ptr++)<<24;
	numfonts += (*ptr++)<<16;
	numfonts += (*ptr++)<<8;
	numfonts += (*ptr++);

	for(i=0; i<numfonts; i++) {
		if(ptr+4 > buf+len){
			printf("File header error - not a valid TrueType collection\n");
			return -1;
		} 

		offset = (*ptr++)<<24;
		offset += (*ptr++)<<16;
		offset += (*ptr++)<<8;
		offset += (*ptr++);

		vptr = buf+offset;
		if(vptr+4 > buf+len){
		printf("Ttc file header error.\n");
		return -1;
	}

		tt->version = (*vptr++)<<24;
		tt->version += (*vptr++)<<16;
		tt->version += (*vptr++)<<8;
		tt->version += (*vptr++);

		printf("\nFont index %d\n",i);

		if(verifyttfotf(buf,buf+offset+4,tt,len))
			return -1;
	}

	return 0;
}



int verifytruetype(unsigned char *buf, truetype_type *tt, int len)
{
	unsigned char *ptr;
	int checksum;
	int i;
	int otto = 0x4f54544f;
	int ttcf = 0x74746366;

	

	


	ptr = buf;
	if(ptr+4 > buf+len){
		printf("File header error - not a valid TrueType font\n");
		return -1;
	}
	tt->version = (*ptr++)<<24;
	tt->version += (*ptr++)<<16;
	tt->version += (*ptr++)<<8;
	tt->version += (*ptr++);

	if(tt->version == 0x00010000 || tt->version == otto) {
		ptr = buf;
		checksum = 0;
		for(i=0; i<len>>2; i++) {
		
			checksum += (*ptr++)<<24;
			checksum += (*ptr++)<<16;
			checksum += (*ptr++)<<8;
			checksum += (*ptr++);
		
		}

		if(checksum != 0xb1b0afba) {
			printf("File checksum error.\n");
		}

		ptr = buf+4;
		verifyttfotf(buf,ptr,tt,len);

	}
	else if(tt->version == ttcf) {
		verifyttc(buf,ptr,tt,len);
	}

	else {
		printf("File header error - not a valid TrueType font\n");
		return -1;
	}

	

	return 0;

}

main(int argc, char *argv[])
{

	FILE *ttfile;
	int len;
	int numbytes;
	unsigned char *buf;

	truetype_type tt;

	if(argc != 2) {
		printf("\n\tUsage: otfontck <[path\\]font_file_name>\n");
		return -1;
	}


	ttfile = fopen(argv[1],"rb");
	if(!ttfile) {
		printf("Error - cannot open file %s\n",argv[1]);
		return -1;
	}


	fseek(ttfile,0,SEEK_END);

	len = ftell(ttfile);

	fseek(ttfile,0,SEEK_SET);

	buf = malloc(len);

	numbytes = fread(buf,1,len,ttfile);

	if(numbytes != len) {
		return -1;
	}


	tt.filelen = numbytes;
   
	fclose(ttfile);

	printf("File: %s\n\n",argv[1]);


	verifytruetype(buf, &tt, len);

	free(buf);
	return 0;
}



